#!/usr/bin/env python2
""" Run the quick integration spec
    This spec is designed to fail at certain points and so because of that it is useful to have a
    wrapper that expects those failures and notes they are really test success.
"""
import os
import shutil
import sys
import subprocess
import time

# Env variables used by mojo
series = 'xenial'
os.environ['MOJO_SERIES'] = series
project = 'quick-integration'
os.environ['MOJO_PROJECT'] = project
workspace = project
os.environ['MOJO_WORKSPACE'] = workspace
stage = 'integration-tests/quick-integration'
os.environ['MOJO_STAGE'] = stage
# TEST: tests that mojo can handle a relative path for the MOJO_SPEC
os.chdir(os.path.dirname(os.path.realpath(__file__)))
# TEST: tests that mojo can handle codetree style source options
os.environ['MOJO_SPEC'] = './;overwrite=true'
root = os.path.expanduser('~/.local/share/mojo')

# A dict of manifests that are expected to fail and their expected failure message
FAIL_MANIFESTS = {}


class TestError(Exception):
    pass


def main(argv=None):
    return_code = 0
    subprocess.check_call(['juju', 'add-model', project])
    try:
        # TEST: project-new
        print("\n**** Running mojo project-new\n")
        subprocess.check_call(['mojo', 'project-new', '--container', 'lxd', project])
        # TEST: project-list
        plist = subprocess.check_output(['mojo', 'project-list']).strip()
        expected = '{}: {}'.format(project, series)
        if expected not in plist.splitlines():
            raise TestError('Error: Expected project-list output to include {} but received {}'.format(expected, plist))

        print("\n**** Running mojo workspace-new\n")
        # TEST: workspace-new
        subprocess.check_call(['mojo', 'workspace-new', os.environ['MOJO_SPEC'], workspace])
        # TEST: workspace-list
        wksp = subprocess.check_output(['mojo', 'workspace-list']).splitlines()[-1]
        if wksp != workspace:
            raise TestError('Error: Expected workspace-list to output {} but received {}'.format(workspace, wksp))

        # Create some secrets, their existence will later be validated during the mojo run
        secret_dir = os.path.join(root, 'LOCAL', project, stage)
        if not os.path.exists(secret_dir):
            os.makedirs(secret_dir)
        with open(os.path.join(secret_dir, 'big-secret'), 'w') as bs:
            bs.write("This quick integration test still isn't that quick.")

        # The primary manifest, everything in here should pass
        print("\n**** Running primary mojo manifest\n")
        # TEST: run of default manifest
        subprocess.check_call(['mojo', 'run'])

        for manifest, expected_error in FAIL_MANIFESTS:
            try:
                print("\n**** Running mojo manifest {}\n".format(manifest))
                run_output = subprocess.check_output(['mojo', 'run', '--manifest-file', manifest])
            except subprocess.CalledProcessError:
                if expected_error in run_output:
                    continue
                else:
                    print('Manifest {} failed with a different error than expected'.format(manifest))
                    print("\tExpected Error:{}\n\nRecevied Error:{}".format(expected_error, run_output))
            else:
                raise TestError('Manifest {} succeeded but it was was expected to fail with {}!'.format(manifest,
                                                                                              expected_error))

        # TEST: Verify the default log exists
        log_file = os.path.join(root, project, series, workspace, 'log', 'mojo.log')
        if not os.path.exists(log_file):
            raise TestError("Error: Default log file {}, doesn't exist.".format(log_file))
        else:
            stats = os.stat(log_file)
            if stats.st_size < 1000:
                raise TestError("Error: Default log file {}, is empty or suspiciously small.".format(log_file))
            if (time.time() - stats.st_mtime) > 60:
                raise TestError("Error: Default log file {}, has not been recently written to.".format(log_file))

        # TEST: workspace-destroy
        subprocess.check_call(['mojo', 'workspace-destroy'])

    except Exception as e:
        return_code = 1
        print('Exception: {}'.format(e))
    finally:
        # TEST: project-destroy
        subprocess.check_call(['mojo', 'project-destroy', project])

        subprocess.check_call(['juju', 'destroy-model', '-y', project])

        shutil.rmtree(secret_dir)

        return return_code


if __name__ == "__main__":
    sys.exit(main())
