# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import os
import shlex
import yaml
import shutil
import logging
import glob


class InvalidCharmSpec(Exception):
    pass


class CharmRepo(object):
    def __init__(self, repo_dir):
        super(CharmRepo, self).__init__()
        self.repo_dir = repo_dir
        if not os.path.exists(repo_dir):
            os.makedirs(repo_dir)
        self.charms = {}

    def charm_metadata(self, charm_dir):
        md_file = os.path.join(charm_dir, 'metadata.yaml')
        with open(md_file) as f:
            return yaml.safe_load(f)

    # XXX: copy defaults to True because of LP#1129319
    def add_charm(self, charm_name, source, series, copy=True):
        md_name = self.charm_metadata(source)['name']
        if md_name in self.charms:
            logging.warn("{} and {} are both named '{}'. It is usually not necessary "
                         "to download a charm more than once - simply use the "
                         "same charm for multiple services.".format(
                             source, self.charms[md_name], md_name))
        self.charms[md_name] = source
        destdir = os.path.join(self.repo_dir, series)
        if not os.path.exists(destdir):
            os.mkdir(destdir)
        dest = os.path.join(destdir, charm_name)
        if os.path.exists(dest):
            shutil.rmtree(dest)
        if copy:
            logging.info("{} => {} (copy)".format(source, dest))
            shutil.copytree(source, dest, symlinks=True)
        else:
            logging.info("{} => {} (symlink)".format(source, dest))
            os.symlink(os.path.relpath(source, os.path.dirname(dest)), dest)

    def build_from_config(self, charm_config, build_dir, series):
        """
        Build a charm repo from a charm config file.

        Lines beginning with "#" are comments. Empty lines are ignored.
        Configuration lines are either

            charm_name

        where the path inside build_dir is expected to be the same as the
        charm name or:

            charm_name path

        where path is relative to build_dir. Charms directories inside the
        repo are always named charm_name.
        """

        with open(charm_config) as f:
            for line in f:
                # Comment lines begin with #, empty lines are ignored
                if line.strip().startswith("#") or len(line.strip()) == 0:
                    continue
                args = shlex.split(line.strip())
                charm_name = args[0]
                if len(args) == 1:
                    # charm path is assumed to be same as the name
                    source = os.path.join(build_dir, args[0])
                elif len(args) == 2:
                    source = os.path.join(build_dir, args[1])
                else:
                    raise InvalidCharmSpec("Unrecognized format for line: {}"
                                           "".format(line))
                self.add_charm(charm_name, source, series)

    def find_charms(self, build_dir):
        """
        Return a list of all charm paths in build_dir.
        """

        charms = []
        for metadata in glob.glob(os.path.join(build_dir,
                                               '*',
                                               'metadata.yaml')):
            if os.path.isfile(metadata):
                charms.append(os.path.dirname(metadata))
        return charms

    def build_all(self, build_dir, series):
        """
        Build a charm repo automatically from all charms in build_dir.
        """

        for source in self.find_charms(build_dir):
            charm_name = os.path.basename(source)
            self.add_charm(charm_name, source, series)
