# Copyright 2014, 2015 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
from __future__ import print_function
import argparse
import errno
import logging
import shutil
import subprocess
import sys
import os

import argcomplete

from textwrap import dedent

import mojo
from .utils import setup_logging, warn_if_root

from .contain import (
    LXCContainer,
    LXDContainer,
    NoContainer,
    NoSuchContainerError,
    )
from .exceptions import ArgumentError, Unprivileged
import mojo.exceptions
import mojo.project
import mojo.workspace


@warn_if_root
def run_phase(args):
    project, workspace = get_project_and_workspace(args)

    # XXX(pjdc): I suspect these are not needed.  The only other place
    # they (and others, but not MOJO_ROOT) are exported is
    # ScriptPhase.run(), which is called if we invoke a script phase.
    os.environ["MOJO_PROJECT"] = project.name
    os.environ["MOJO_ROOT"] = project.mojo_root
    os.environ["MOJO_WORKSPACE"] = workspace.name

    phase = mojo.phase.build(args.subparser_name, **args.options)
    phase.run(project, workspace, args.stage)


def show_deploy(args):
    "Show rendered juju-deployer configuration"
    project, workspace = get_project_and_workspace(args)
    phase = mojo.phase.build('deploy', **args.options)
    phase.show(project, workspace, args.stage)


def diff_deploy(args):
    """
    Show juju-deployer diff. Generate a delta between a
    configured deployment and a running environment
    """
    project, workspace = get_project_and_workspace(args)
    phase = None

    if args.apply_config_changes:
        phase = mojo.phase.build('apply-config-changes', **args.options)

    if args.preview_config_changes:
        phase = mojo.phase.build('preview-config-changes', **args.options)

    if not phase:
        phase = mojo.phase.build('deploy', **args.options)

    phase.diff(project, workspace, args.stage)


def exit_with_usage(parser, message):
    print(message + "\n", file=sys.stderr)
    parser.print_help(file=sys.stderr)
    sys.exit(1)


class OptionParser(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        options = {}

        for option in values:
            try:
                (okey, oval) = option.split('=', 1)
            except ValueError:
                exit_with_usage(
                    parser, "ERROR: Options must be key=value pairs.")

            if okey in options:
                exit_with_usage(
                    parser, 'ERROR: Option {}={} would override '
                    'previously supplied option {}={}.'.format(
                        okey, oval, okey, options[okey]))

            options[okey] = oval

        setattr(namespace, self.dest, options)


# -----------------------
# helpers
# -----------------------
def get_project_and_workspace(args):
    project_name = get_project_name(args, use_spec=True)
    spec_url, workspace = get_spec_url_and_workspace(args)
    project = mojo.project.Project(project_name, args.series, args.mojo_root)
    workspace = project.create_workspace(spec_url, workspace, stage=args.stage, logfile=args.logfile)
    return project, workspace


def get_project_name(args, use_spec=False):
    """Return the project name from args or the environment.

    Priority for setting project: command line switch then
    MOJO_PROJECT environment variable

    If use_spec is true and the above priority settings are
    missing, then the project will be guessed from spec_url.
    A warning will be printed to this effect.

    Raises ArgumentError if no project can be determined.
    """

    # command line switch, --project
    project_name = args.project
    # MOJO_PROJECT environment variable
    if not project_name:
        project_name = os.environ.get("MOJO_PROJECT")
    # Guess based on spec_url and warn
    if not project_name and use_spec:
        spec_url, workspace = get_spec_url_and_workspace(args)
        project_name = spec_url.rsplit("/")[-1]
        if not project_name:
            project_name = args.spec_url.rsplit("/")[-2]
        logging.warn("Guessing the project is {}. Use --project or set "
                     "MOJO_PROJECT to be more explicit.".format(project_name))
    # Fail gracefully
    if not project_name:
        raise ArgumentError("Project setting missing. Use --project or set "
                            "MOJO_PROJECT")
    return project_name


def get_spec_url_and_workspace(args):
    """Return the spec_url and workspace from args or the
    environment.

    Returns spec_url and workspace.

    Raises ArgumentError if either spec_url or workspace
    cannot be determined.
    """
    spec_url = args.spec_url or os.environ.get("MOJO_SPEC", ".")
    workspace = args.workspace or os.environ.get("MOJO_WORKSPACE")

    # Note: this won't catch links to paths inside the mojo_root
    if args.mojo_root is not None and os.path.commonprefix([args.mojo_root, spec_url]) == args.mojo_root:
        raise ArgumentError("MOJO_SPEC can't be a subdirectory of MOJO_ROOT")

    if not spec_url or not workspace or spec_url == workspace:
        raise ArgumentError("Missing spec_url and/or workspace. Set "
                            "MOJO_SPEC and/or MOJO_WORKSPACE")

    return (spec_url, workspace)


def get_container(args):
    """Return the container from args or the environment.
    Returns 'LXCContainer' by default.
    Currently this is only required at project creation time.
    Stateful project information will be gathered on subsequent
    mojo commands. If this behavior changes, the defaults below
    may need to change.
    """
    if args.container == 'lxc':
        return LXCContainer
    elif args.container == 'lxd':
        return LXDContainer
    elif args.container == 'containerless':
        return NoContainer
    else:
        raise NoSuchContainerError("{} is an unsupported container type."
                                   "".format(args.container))


# -----------------------
# commands called from wrapper scripts
# -----------------------
def create_project(args):
    "Set up a Mojo project and LXC container"
    project = mojo.project.Project(args.name, args.series, args.mojo_root,
                                   container_class=get_container(args))
    project.create()


def destroy_project(args):
    try:
        if args.series:
            series_list = (args.series,)
        else:
            series_list = mojo.project.Project.list_series(args.mojo_root, args.name)

        projects = [mojo.project.Project(args.name, series, args.mojo_root) for series in series_list]

        for project in projects:
            logging.info("Destroying project: {}, series: {}"
                         "".format(project.name, project.series))
            project.delete()

    except mojo.project.ProjectNameError:
        logging.error('Invalid project name: {}'.format(args.name))
        sys.exit(1)
    except mojo.project.ProjectDoesNotExistError:
        logging.error('Invalid project: {}'.format(args.name))
        sys.exit(1)
    except mojo.project.ProjectCreationError:
        logging.exception('Invalid project: {}'.format(args.name))
        sys.exit(1)


# -----------------------
# commands
# -----------------------
@warn_if_root
def call_project_new(args):
    """Create a new project. Call via sudo if we're creating a privileged container"""
    cmd = ("mojo-project-new", "--series", args.series, "--container", args.container, '--log_level', args.log_level, args.name)
    if args.mojo_root:
        cmd = cmd + ("--mojo_root", args.mojo_root)
    if args.container == "lxc":
        cmd = ("sudo", ) + cmd
        if os.path.realpath(sys.argv[0]).startswith('/snap'):
            raise Unprivileged("Only LXD containers and containerless projects are supported with a snap install.")
    logging.debug("Running command: {}".format(" ".join(cmd)))
    subprocess.check_call(cmd)


@warn_if_root
def call_project_destroy(args):
    """Destroy a project. Call via sudo if it has a privileged container"""
    cmd = ("mojo-project-destroy", '--log_level', args.log_level)
    if args.mojo_root:
        cmd = cmd + ("--mojo_root", args.mojo_root)
    cmd = cmd + ("--series", args.series, args.name)
    project = mojo.project.Project(args.name, args.series, args.mojo_root)
    if project.project_info.get("container_class") == "lxc":
        cmd = ("sudo", ) + cmd
        if os.path.realpath(sys.argv[0]).startswith('/snap'):
            raise Unprivileged("Only LXD containers and containerless projects are supported with a snap install.")
    logging.debug("Running command: {}".format(" ".join(cmd)))
    subprocess.call(cmd)


def list_projects(args):
    map(print, mojo.project.Project.list(args.mojo_root, only_names=args.names))


@warn_if_root
def create_workspace(args):
    "Set up workspace and fetch a copy of the spec"
    project = mojo.project.Project(get_project_name(args), args.series, args.mojo_root)
    spec_url, workspace = get_spec_url_and_workspace(args)

    try:
        project.create_workspace(spec_url, workspace, stage=args.stage, logfile=args.logfile)
    except OSError:
        print("Are you sure the project exists?", file=sys.stderr)
        if args.blunt:
            raise
        sys.exit(2)
    except AttributeError:
        logging.error("Unable to branch {}".format(args.spec_url))
        logging.error("Please check the branch exists and you have permissions to "
                      "access it")
        # Now we need to clean up the workspace
        project.delete_workspace(workspace)
        sys.exit(2)
    logging.info("Workspace created.")


@warn_if_root
def destroy_workspace(args):
    project_name = get_project_name(args)

    try:
        project = mojo.project.Project(project_name, args.series, args.mojo_root)
        project.workspace(args.workspace)
    except mojo.workspace.NoSuchWorkspaceError:
        logging.error(
            'No such workspace (project: {project}; series: {series}): {workspace}'.format(
                project=project_name,
                series=args.series,
                workspace=args.workspace,
                ))
        sys.exit(1)
    except mojo.project.ProjectNameError:
        logging.error('Invalid project name: {}'.format(project_name))
        sys.exit(1)
    except mojo.project.ProjectDoesNotExistError:
        logging.error('Invalid project: {}'.format(project_name))
        sys.exit(1)
    except mojo.project.ProjectCreationError:
        logging.exception('Invalid project: {}'.format(project_name))
        sys.exit(1)

    project.delete_workspace(args.workspace)


def list_workspaces(args):
    project = mojo.project.Project(get_project_name(args), args.series, args.mojo_root)
    map(print, project.workspace_list())


def new_spec(args):
    spec_template = '/usr/share/mojo/contrib/spec-template'
    if not os.path.exists(spec_template):
        logging.error('Unable to find spec template at {}'.format(spec_template))
        sys.exit(1)
    shutil.copytree(spec_template, args.dir)
    logging.info('New spec at {}'.format(args.dir))


def display_help(args):
    "Display help for sub-commands"
    if not args.command:
        args.parser.print_help()
    else:
        args.parser.parse_args([args.command, '-h'])
    sys.exit(0)


@warn_if_root
def run_from_manifest(args):
    project_name = get_project_name(args, use_spec=True)
    spec_url, workspace = get_spec_url_and_workspace(args)
    project = mojo.project.Project(project_name, args.series, args.mojo_root)
    workspace = project.create_workspace(spec_url, workspace,
                                         stage=args.stage,
                                         manifest_file=args.manifest_file,
                                         logfile=args.logfile)
    manifest = workspace.spec.manifest
    manifest.run(project, workspace, args.stage, args.interactive)


class MojoArgumentParser(argparse.ArgumentParser):
    """
    An ArgumentParser that silently accepts long options containing
    underscores and converts them to the corresponding option using hypens.

    ie. --foo_bar is converted to --foo-bar. Non-option arguments
    are left untouched.
    """
    def add_argument(self, *args, **kwargs):
        # Replace _ with - to create a consistent cli, no matter
        # what the callsite chose. We don't need to worry about
        # the 'dest' argument as argparse already munges that.
        args = list(args)
        for i, arg in enumerate(args):
            if arg.startswith('--'):
                args[i] = '--' + arg[2:].replace('_', '-')
        return super(MojoArgumentParser, self).add_argument(*args, **kwargs)

    def parse_known_args(self, args=None, namespace=None):
        if args is None:
            # Default to the system args
            args = sys.argv[1:]
        args = list(args)
        for i, arg in enumerate(args):
            if arg == '--':
                break  # Everything after -- is passed through as is.
            if arg.startswith('--'):
                # --foo_bar is converted to --foo-bar.
                args[i] = '--' + arg[2:].replace('_', '-')
        return super(MojoArgumentParser, self).parse_known_args(args,
                                                                namespace)


def create_arg_parser():
    # main program parser
    ap = MojoArgumentParser()
    ap.add_argument("--blunt", action="store_true",
                    help="Handle fewer exceptions")
    ap.add_argument("--break_everything", action="store_true",
                    help="Allow running as root, possibly breaking things")
    ap.add_argument("--log_level", "-l", default='INFO',
                    help="Log level. Defaults to INFO")
    ap.add_argument("--logfile", "-L", default=os.environ.get('MOJO_LOGFILE'),
                    help="File to log to in addition to stdout. Overrides environment "
                         "variable MOJO_LOGFILE. Defaults to MOJO_WORKSPACE/log/mojo.log")
    ap.add_argument("-r", "--mojo_root", default=os.environ.get('MOJO_ROOT'),
                    help="Root directory that mojo works out of Overrides the MOJO_ROOT environment variable. "
                         "Defaults to /srv/mojo for projects using 'lxc' containers and to "
                         " ~/.local/share/mojo for 'lxd' and 'containerless' projects.")
    ap.add_argument('--version', action='version',
                    version=mojo.__version__)

    # subcommand parser
    sp = ap.add_subparsers(title="subcommands", dest="subparser_name")

    # Common subcommand args, set here in one place and then reuse in subcommands as needed
    project_args = ("-p", "--project")
    project_kwargs = {'help': "Project name. Overrides the MOJO_PROJECT environment variable."}
    series_args = ("-s", "--series")
    series_kwargs = {'default': os.environ.get('MOJO_SERIES'),
                     'help': "Distro series. Overrides the MOJO_SERIES environment variable."}
    stage_args = ("--stage", )
    stage_kwargs = {'default': os.environ.get('MOJO_STAGE', 'devel'),
                    'help': "Deployment stage. Overrides MOJO_STAGE environment variable (default: devel)"}
    workspace_args = ("-w", "--workspace")
    workspace_kwargs = {'default': os.environ.get('MOJO_WORKSPACE'),
                        'help': "Workspace name. Overrides the MOJO_WORKSPACE environment variable."}

    # options parser for phase commands
    phaseopts = MojoArgumentParser(add_help=False)
    phaseopts.add_argument(*stage_args, **stage_kwargs)
    phaseopts.add_argument("--options",
                           help="Phase options. i.e. config=CONFIG_FILENAME",
                           nargs="*", metavar="key=value", action=OptionParser,
                           default={})
    phaseopts.add_argument(*project_args, **project_kwargs)
    phaseopts.add_argument(*workspace_args, **workspace_kwargs)
    phaseopts.add_argument(*series_args, **series_kwargs)
    phaseopts.add_argument("--spec-url", default=None,
                           help='Location of the Mojo spec. Overrides the '
                                'MOJO_SPEC environment variable. If MOJO_SPEC '
                                'is unset, defaults to ".".')

    # project-new subcommand - creates a new mojo project
    prjnewcmd = sp.add_parser("project-new",
                              description="Create a new Mojo project and container")
    prjnewcmd.add_argument(*series_args, required=False, **series_kwargs)
    prjnewcmd.add_argument("name", nargs='?', default=os.environ.get('MOJO_PROJECT'),
                           help="Base name of the project ([a-zA-Z0-9._]*)")
    prjnewcmd.add_argument('-c', "--container", default=os.environ.get('MOJO_CONTAINER', 'lxc'),
                           choices=['lxc', 'lxd', 'containerless'],
                           help="Container. Options 'lxc', 'lxd' or 'containerless'. "
                                "Defaults to 'lxc'")
    prjnewcmd.set_defaults(func=call_project_new)

    # project-destroy subcommand - deletes a mojo project
    prjdelcmd = sp.add_parser("project-destroy",
                              description="Destroy a Mojo project and container")
    prjdelcmd.add_argument("name",
                           help="Base name of the project ([a-zA-Z0-9._]*)")
    prjdelcmd.add_argument(*series_args, **series_kwargs)
    prjdelcmd.set_defaults(func=call_project_destroy)

    # project-list
    prjlistcmd = sp.add_parser("project-list",
                               description="List existing projects, found at the mojo root")
    prjlistcmd.add_argument("-n", "--names", action="store_true",
                            help="List only project names, without series")
    prjlistcmd.set_defaults(func=list_projects)

    # workspace-new subcommand - creates a new workspace for a specific spec
    # revision
    wsnewcmd = sp.add_parser("workspace-new",
                             description="Initialize a mojo workspace")
    wsnewcmd.add_argument("spec_url", nargs='?', default=None,
                          help='Location of the Mojo spec. Overrides the '
                               'MOJO_SPEC environment variable. If MOJO_SPEC '
                               'is unset, defaults to ".".')
    wsnewcmd.add_argument("workspace", nargs='?', default=None, help=workspace_kwargs['help'])
    wsnewcmd.add_argument(*project_args, **project_kwargs)
    wsnewcmd.add_argument(*series_args, **series_kwargs)
    wsnewcmd.add_argument(*stage_args, **stage_kwargs)
    wsnewcmd.set_defaults(func=create_workspace, parser=wsnewcmd)

    # workspace-destroy subcommand - deletes a mojo project
    wsdelcmd = sp.add_parser("workspace-destroy",
                             description="Destroy a mojo workspace")
    wsdelcmd.add_argument(*project_args, **project_kwargs)
    wsdelcmd.add_argument(*series_args, **series_kwargs)
    wsdelcmd.add_argument(*workspace_args, **workspace_kwargs)
    wsdelcmd.set_defaults(func=destroy_workspace)

    # workspace-list subcommand
    wslistcmd = sp.add_parser("workspace-list",
                              description="List workspaces for a project")
    wslistcmd.add_argument(*project_args, **project_kwargs)
    wslistcmd.add_argument(*series_args, **series_kwargs)
    wslistcmd.set_defaults(func=list_workspaces)

    # builddeps subcommand - install build dependencies for project
    depscmd = sp.add_parser("builddeps", description="builddeps",
                            parents=[phaseopts])
    depscmd.set_defaults(func=run_phase)

    # collect subcommand - collects code branches to build directory
    clctcmd = sp.add_parser("collect",
                            description="Collect resources for build",
                            parents=[phaseopts])
    clctcmd.set_defaults(func=run_phase)

    # charm-audit subcommand - runs a charm audit
    charmauditcmd = sp.add_parser("charm-audit",
                                  description="Run a charm audit",
                                  parents=[phaseopts])
    charmauditcmd.set_defaults(func=run_phase)

    # build subcommand - run project build script inside container
    buildcmd = sp.add_parser("build",
                             description="Execute project build script",
                             parents=[phaseopts])
    buildcmd.set_defaults(func=run_phase)

    # repo subcommand - assembles charms into a repository
    repocmd = sp.add_parser("repo", description="Assemble a charm repository",
                            parents=[phaseopts])
    repocmd.set_defaults(func=run_phase)

    # sleep subcommand - sleeps for n seconds
    sleepcmd = sp.add_parser("sleep", description="Sleep for time in seconds",
                             parents=[phaseopts])
    sleepcmd.set_defaults(func=run_phase)

    # secrets subcommand - secretss for n seconds
    secretscmd = sp.add_parser("secrets",
                               description="Copy secrets from staged "
                               "location to workspace local",
                               parents=[phaseopts])
    secretscmd.set_defaults(func=run_phase)

    # script subcommand - run an arbitrary script inside container
    scriptcmd = sp.add_parser("script", description="Execute a project script",
                              parents=[phaseopts])
    scriptcmd.set_defaults(func=run_phase)

    # bundle subcommand - run a juju deploy using bundles
    bundlecmd = sp.add_parser("bundle",
                              description="Deploy applications and relations using juju bundles",
                              parents=[phaseopts])
    bundlecmd.set_defaults(func=run_phase)

    # deploy subcommand - run a juju-deployer session
    deploycmd = sp.add_parser("deploy",
                              description="Deploy services and/or create "
                              "relations",
                              parents=[phaseopts])
    deploycmd.add_argument("--ignore", action="store_true", default=False,
                           help="Exit 0 even if service deployment fails")
    deploycmd.set_defaults(func=run_phase)

    # deploy-show subcommand
    deployshowcmd = sp.add_parser("deploy-show",
                                  description="Show rendered juju-deployer "
                                              "configuration",
                                  parents=[phaseopts])
    deployshowcmd.set_defaults(func=show_deploy)

    # deploy-diff subcommand
    deploydiffcmd = sp.add_parser("deploy-diff",
                                  description="Show juju-deployer diff. "
                                              "Generate a delta between a "
                                              "configured deployment and a "
                                              "running environment.",
                                  parents=[phaseopts])

    configapplicationgroup = deploydiffcmd.add_mutually_exclusive_group()
    configapplicationgroup.add_argument(
        "--apply-config-changes",
        action="store_true", default=False,
        help="Apply config differences")
    configapplicationgroup.add_argument(
        "--preview-config-changes",
        action="store_true", default=False,
        help="Preview config application")

    deploydiffcmd.set_defaults(func=diff_deploy)

    # volumes subcommand - mount volumes
    volcmd = sp.add_parser("volumes", description="Mount volumes",
                           parents=[phaseopts])
    volcmd.set_defaults(func=run_phase)

    # verify subcommand - verify that a deployment was successful
    verifycmd = sp.add_parser("verify",
                              description="Verify that the deployment is "
                              "working",
                              parents=[phaseopts])
    verifycmd.set_defaults(func=run_phase)

    # juju-check-wait subcommand - wait until juju environment reaches steady state
    jujucheckwaitcmd = sp.add_parser("juju-check-wait",
                                     description="Wait until juju environment "
                                                 "reaches steady state.",
                                     parents=[phaseopts])
    jujucheckwaitcmd.set_defaults(func=run_phase)

    # nagios-check subcommand - run all nagios checks in an environment
    nagioscheckcmd = sp.add_parser("nagios-check",
                                   description="Run all nagios checks in the environment",
                                   parents=[phaseopts])
    nagioscheckcmd.set_defaults(func=run_phase)

    # run subcommand - run an entire spec per its manifest
    runcmd = sp.add_parser("run",
                           description="Run an entire deployment from start "
                           "to finish")
    runcmd.add_argument("spec_url", nargs='?', default=None,
                        help='Location of the Mojo spec. Overrides the '
                             'MOJO_SPEC environment variable. If MOJO_SPEC '
                             'is unset, defaults to ".".')
    runcmd.add_argument("workspace", nargs='?', default=None, help=workspace_kwargs['help'])
    runcmd.add_argument("-i", "--interactive", action="store_true", default=False,
                        help="Run the manifest in interactive mode, prompting for each phase before running it")
    runcmd.add_argument("-m", "--manifest-file", default="manifest",
                        help="The particular manifest file to run")
    runcmd.add_argument(*series_args, **series_kwargs)
    runcmd.add_argument(*stage_args, **stage_kwargs)
    runcmd.add_argument(*project_args, **project_kwargs)
    runcmd.set_defaults(func=run_from_manifest)

    # new-spec subcommand
    newspeccmd = sp.add_parser("new-spec", description="Create a new spec template at the given path.")
    newspeccmd.add_argument("dir", help="Directory for the new spec, absolute or relative to current working dir")
    newspeccmd.set_defaults(func=new_spec)

    # help subcommand - get help for other subcommands
    helpcmd = sp.add_parser("help", description="Interactive help")
    helpcmd.add_argument("command", nargs='?', default=None,
                         help="Get help for this sub-command.")
    helpcmd.set_defaults(func=display_help, parser=ap)

    return ap


def run_with_args(args, ap=None):
    if hasattr(args, "cache_ttl") and args.cache_ttl is not None:
        mojo.CacheDisk.ttl(args.cache_ttl)
    try:
        args.func(args)
    except ArgumentError as e:
        logging.error(e.message)
        if ap is not None:
            ap.parse_args([args.subparser_name, '-h'])
        sys.exit(1)
    except Unprivileged:
        logging.error("This sub-command, {}, must be run as root"
                      "".format(args.subparser_name))
        sys.exit(1)
    except (mojo.juju.status.JujuStatusError,
            mojo.juju.JujuWaitException,
            mojo.phase.CollectPhaseException,
            mojo.phase.NagiosCheckPhaseException,
            mojo.phase.StopOnPhaseException,
            mojo.phase.VerifyPhaseException,
            mojo.phase.JujuDeployerException) as e:
        logging.error(e)
        sys.exit(1)
    except mojo.phase.ScriptPhaseException as e:
        logging.error(e.output)
        sys.exit(1)
    except mojo.exceptions.ConfigNotFoundException as e:
        logging.error(e.message)
        print("Searched:", file=sys.stderr)
        for path in e.search_list:
            print("  {}".format(path), file=sys.stderr)
        sys.exit(1)
    except (mojo.exceptions.InvalidManifestException,
            mojo.phase.FileDoesNotExistException,
            mojo.phase.InvalidStopOnReturnCodeException,
            mojo.phase.InvalidYAMLException) as e:
        logging.error(e.message)
        sys.exit(1)
    except subprocess.CalledProcessError:
        logging.exception("Error calling external process")
        sys.exit(1)
    except IOError as ioe:
        # Try to diagnose problems with /var/lib/lxc;
        # https://bugs.launchpad.net/mojo/+bug/1346873
        filename = ioe.filename
        if not (filename is not None and filename.startswith('/var/lib/lxc') and
                os.path.basename(filename).startswith('config')):
            raise
        if ioe.errno != errno.EACCES:
            raise

        print('ERROR: Unable to read container configuration '
              'from {}'.format(filename), file=sys.stderr)

        try:
            os.listdir('/var/lib/lxc')
        except OSError as ose:
            if ose.errno == errno.EACCES:
                print(dedent(
                    '''\
                    ERROR: /var/lib/lxc is not readable; consider one of the following:

                    Option 1:
                        sudo chgrp mojo /var/lib/lxc
                        sudo chmod 0750 /var/lib/lxc

                    Option 2:
                        sudo chmod 0755 /var/lib/lxc'''),
                      file=sys.stderr)
                sys.exit(1)

        try:
            container = os.path.dirname(filename)
            os.listdir(container)
        except OSError as ose:
            # The problem isn't /var/lib/lxc; it's the container!
            if ose.errno == errno.EACCES:
                # We know it's not /var/lib/lxc; that was checked above.
                print(dedent(
                    '''\
                    ERROR: {container} exists but is not readable; consider doing the following:

                        sudo chmod 0755 {container}''').format(
                            container=container),
                      file=sys.stderr)
                sys.exit(1)
            if ose.errno == errno.ENOENT:
                # Mojo checks this itself, so this would be very strange.
                print('ERROR: {} does not exist; was the Mojo project '
                      'created properly?!'.format(container), file=sys.stderr)
                sys.exit(1)

        # Some things are beyond even our powers to diagnose.
        raise

    except Exception as e:
        logging.exception("Unknown error")
        sys.exit(1)


UNWANTED_ENVIRONMENT_VARIABLES = [
    # We expect juju-deployer to look in the cwd,
    # but setting $JUJU_REPOSITORY overrides that.
    'JUJU_REPOSITORY',
    ]


def sanitize_environment():
    for variable in UNWANTED_ENVIRONMENT_VARIABLES:
        if variable in os.environ:
            logging.warning(
                'Removing ${} from the environment.'.format(variable)
                )
            del(os.environ[variable])


# -----------------------
# main
# -----------------------


def main():
    ap = create_arg_parser()
    argcomplete.autocomplete(ap)
    args = ap.parse_args()
    setup_logging(args.log_level)

    sanitize_environment()

    run_with_args(args, ap)


if __name__ == "__main__":
    sys.exit(main())
