# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
#
# check for LXC container for workspace
# create LXC container for workspace
# execute scripts inside of container
# enable or disable networking inside the container

import ast
import logging
import os
import platform
import pwd
import shlex
import shutil
import subprocess
import sys
import tempfile
import time
import yaml
from contextlib import contextmanager
from distutils.version import LooseVersion
from .utils import is_config_line, needs_root
try:
    import pylxd
    from pylxd.exceptions import LXDAPIException
except ImportError:
    PYLXD_AVAILABLE = False
else:
    PYLXD_AVAILABLE = True

UBUNTU_RELEASES = {'precise': '12.04',
                   'trusty': '14.04',
                   'utopic': '14.10',
                   'vivid': '15.04',
                   'wily': '15.10',
                   'xenial': '16.04',
                   'yakkety': '16.10',
                   'zesty': '17.04',
                   'artful': '17.10',
                   'bionic': '18.04',
                   'cosmic': '18.10',
                   'disco': '19.04',
                   'eoan': '19.10',
                   'focal': '20.04',
                   }

ARCHITECTURES = {'32bit': 'i386',
                 '64bit': 'amd64'
                 }


logging.getLogger("ws4py").setLevel(logging.WARNING)  # Suppress info logging from ws4py used by pylxd


class ContainerTimeout(Exception):
    pass


class NoSuchContainerError(Exception):
    pass


class InvalidConfigurationError(Exception):
    pass


class InvalidUbuntuSeries(Exception):
    pass


# Decorator for methods that require a defined container
def ensure_defined(method):
    def wrapped(self, *args, **kwargs):
        if not self.defined:
            raise NoSuchContainerError(self.container_name)
        return method(self, *args, **kwargs)
    return wrapped


class Container(object):
    def __init__(self, name, series, ensure_exists=False):
        self.name = name
        self.series = series
        self.container_name = "{}.{}".format(name, series)
        if ensure_exists and not self.defined:
            raise NoSuchContainerError(self.container_name)
        if self.defined:
            self.load_config()
        self.utils = ContainerUtils(self)
        if not UBUNTU_RELEASES.get(self.series):
            raise InvalidUbuntuSeries("{} is not a valid Ubuntu series. It "
                                      "may no longer be supported or it may "
                                      "be new and needs to be added to mojo."
                                      "".format(self.series))

    def load_config(self):
        pass

    @contextmanager
    def running_container(self, network=True, keep_running=False):
        running = self.running
        if not running:
            self.start(network=network)
        try:
            yield
        finally:
            if not running and not keep_running:
                self.stop()

    def create(self):
        raise NotImplemented

    def destroy(self, force=False):
        raise NotImplemented

    def exists(self):
        raise NotImplemented

    def is_setup(self):
        raise NotImplemented

    def put(self, src_file, dst_file, uid=0, gid=0, mode=0o644):
        raise NotImplemented

    def start(self):
        raise NotImplemented

    def stop(self):
        raise NotImplemented

    def run(self, command, env=None, user=None, network=True, stdout=None,
            stderr=None):
        raise NotImplemented

    @property
    def conf(self):
        return None

    @property
    def defined(self):
        return False
    # TODO Clarify the class, is there a distinction between exists and defined? If so what what is is_setup()

    @property
    def running(self):
        return False

    @property
    def container_path(self):
        return None

    @property
    def rootfs(self):
        return None


class LXDContainer(Container):
    def __init__(self, name, series, ensure_exists=False):
        if not PYLXD_AVAILABLE:
            raise ImportError("Packages python-pylxd, python-ws4py and lxd are required for "
                              "LXD support and is only supported on xenial and later.")
        self.lxd = pylxd.Client()  # Assumes LXD on localhost
        self._container = None

        Container.__init__(self, name, series, ensure_exists)

        # No periods allowed in LXD container names
        if name.find(".") != -1:
            raise InvalidConfigurationError('"." in not a valid character for LXD container names')
        self.container_name = "{}-{}".format(name, series)

        self.container_class = 'lxd'
        self.container_root = os.path.expanduser(
            "~/.local/share/lxc/{}/rootfs".format(self.container_name))

        self.device_name = "mojo-project-series-root"
        self.timeout = 60

        self._lxd_ver = LooseVersion(pylxd.__version__)

    def _at_least_lxd_ver(self, ver):
        return self._lxd_ver >= LooseVersion(ver)

    def _check_mount(self):
        if self.device_name in self.container.devices.keys():
            return True
        return False

    def _check_state(self):
        return self.container.state().status

    @property
    def _container_uid_offset(self):
        """Returns the container id as an integer"""
        vsm = self.container.config['volatile.last_state.idmap']
        return ast.literal_eval(vsm.replace('true', 'True').replace('false', 'False'))[0]["Hostid"]

    def _disable_network(self):
        if not self._is_network_enabled():
            return

        self.container.expanded_devices['eth0'] = {'type': 'none'}
        self.container.save()

    def _enable_network(self):
        if self._is_network_enabled():
            return
        self.container.expanded_devices['eth0'] = {'type': 'nic', 'nictype': 'bridged', 'parent': 'lxdbr0'}
        self.container.save()

    def _is_network_enabled(self):
        if 'eth0' in self.container.expanded_devices and self.container.expanded_devices['eth0']['type'] != 'none':
            return True
        return False

    def _setup_mount(self):
        if self._check_mount():
            return

        if not os.path.exists(self.container_root):
            raise InvalidConfigurationError("Host path {} does not exist.".format(self.container_root))

        self.container.devices[self.device_name] = {'path': self.container_root,
                                                    'source': self.container_root,
                                                    'type': 'disk'}
        logging.info("Adding device {} to LXC container {}".format(self.device_name, self.container_name))
        self.container.save()
        uid_in_container = self._container_uid_offset
        cmd = ["setfacl", "-Rdm" "u:{}:rwX".format(uid_in_container), self.container_root]
        logging.info("Setting acls inside container: {}".format(" ".join(cmd)))
        subprocess.check_output(cmd)

    def _wait(self, state='Running'):
        logging.info("Waiting for LXD container to get to state '{}'" "".format(state))
        wait_time = 2
        max_waits = self.timeout / wait_time
        wait_num = 0
        while True:
            if self._check_state() == state:
                return
            if wait_num == max_waits:
                raise ContainerTimeout("Timeout ({}s) waiting for state {}".format(self.timeout, state))
            wait_num += 1
            time.sleep(wait_time)

    @property
    def conf(self):
        # XXX: this is needed for project creation currently, but doesn't really apply to LXD
        return self.container_root

    @property
    def container(self):
        if self._container is None:
            try:
                c = self.lxd.containers.get(self.container_name)
            except LXDAPIException:
                return None
            self._container = c

        try:
            # If a container has been destroyed state() will fail
            self._container.state()
        except (LXDAPIException, AttributeError):  # Sometimes in testing it fails with AttributeError
            self._container = None

        return self._container

    @property
    def defined(self):
        return self.exists()

    @property
    def running(self):
        return self._check_state() == "Running"

    def create(self):
        if self.defined:
            raise InvalidConfigurationError("Duplicate container: {}"
                                            "".format(self.container_name))
        dist = platform.dist()[0].lower()
        arch = ARCHITECTURES.get(platform.architecture()[0])
        lxd_alias = 'mojo-{}-{}-{}'.format(dist, self.series, arch)

        # Ensure we have an image alias defined matching what we need
        lxd_aliases = []
        for image in self.lxd.images.all():
            image_aliases = [alias['name'] for alias in image.aliases]
            lxd_aliases.extend(image_aliases)
        if lxd_alias not in lxd_aliases:
            logging.info("Retrieving image from ubuntu:{}.".format(self.series))
            try:
                image = self.lxd.images.create_from_simplestreams("https://cloud-images.ubuntu.com/releases",
                                                                  self.series, public=True, auto_update=True)
                image.add_alias(lxd_alias, "")
            except LXDAPIException:
                raise NoSuchContainerError("No LXD image for series {} found from "
                                           "https://cloud-images.ubuntu.com/release".format(self.series))
        else:
            logging.info("Using mojo container image {}.".format(lxd_alias))

        logging.info("Launching container '{}'".format(self.container_name))
        try:
            config = {'name': self.container_name, 'source': {'type': 'image', 'alias': lxd_alias}}
            self.lxd.containers.create(config, wait=True)
        except LXDAPIException:
            logging.error("There was a problem creating the container")
            raise

        if os.path.exists(self.container_root):
            self._setup_mount()

    @ensure_defined
    def destroy(self, force=False):
        if self.running:
            self.stop()
        self.container.delete(wait=True)

    def exists(self):
        if self.container is None:
            return False
        return True

    def is_setup(self):
        if self.container is None:
            return False

        return self._check_mount()

    @ensure_defined
    def put(self, src_file, dst_file, uid=0, gid=0, mode=0o644):
        with open(src_file, 'r') as src:
            data = src.read()
        self.container.files.put(dst_file, data)
        if self._at_least_lxd_ver('2.2.0'):
            _, _, err = self.container.execute(['chown', str(uid) + ':' + str(gid), dst_file])
            _, _, err2 = self.container.execute(['chmod', '%03o' % mode, dst_file])
        else:
            _, err = self.container.execute(['chown', str(uid) + ':' + str(gid), dst_file])
            _, err2 = self.container.execute(['chmod', '%03o' % mode, dst_file])
        if len(err) + len(err2) > 0:
            logging.error("Error setting ownership and/or permissions:{} {}".format(err, err2))

    @ensure_defined
    def run(self, command, env=None, user=None, network=True, stdout=None, stderr=None):
        """ Run commands in the LXD environment.
            Note: user is not used for LXD the command is run via the LXD API.
        """
        if not command:
            logging.warn("Refusing to execute empty command")
            return -1

        if stdout is None:
            stdout = sys.stdout
        if stderr is None:
            stderr = sys.stderr

        command = shlex.split(command)
        if isinstance(env, dict):
            env_vars = ["env"] + ["=".join((k, v)) for k, v in env.items()]
            command = env_vars + command

        self.start(network=network)

        logging.info("Running command in container '{}': {}"
                     "".format(self.container_name, ' '.join(command)))
        if self._at_least_lxd_ver('2.2.0'):
            _, output, err = self.container.execute(command)
        else:
            output, err = self.container.execute(command)
        stdout.write(output)
        stderr.write(err)
        if output.strip():
            logging.info(output + err)

        if len(err) > 0:
            return 1
        return 0

    @ensure_defined
    def start(self, network=True):
        if self.running:
            if not self._is_network_enabled() and network:
                self.stop()
                self._enable_network()
                self.start()
            elif self._is_network_enabled() and not network:
                self.stop()
                self._disable_network()
                self.start()
            else:
                logging.info("LXD container '{}' is already running"
                             "".format(self.container_name))
        else:
            if not self._is_network_enabled() and network:
                self._enable_network()
            elif self._is_network_enabled() and not network:
                self._disable_network()
            logging.info("Starting LXD container '{}'".format(self.container_name))
            self.container.start(timeout=60)
            self._wait("Running")

            # even after startup networking takes a moment to be ready
            if network:
                wait_time = 1
                while wait_time < 60:
                    time.sleep(1)
                    for addr in self.container.state().network['eth0']['addresses']:
                        if addr['family'] == 'inet' and 'address' in addr:
                            break
                    wait_time += 1

    @ensure_defined
    def stop(self, wait=True):
        if not self.running:
            logging.info("LXD container '{}' is already stopped".format(self.container_name))
        else:
            logging.info("Stopping LXD to container '{}'".format(self.container_name))
            self.container.stop(timeout=60)
            self._wait("Stopped")


class LXCContainer(Container):
    def __init__(self, name, series, ensure_exists=False):
        Container.__init__(self, name, series, ensure_exists)
        self.container_class = 'lxc'
        self.container_root = "/var/lib/lxc/{}/rootfs" \
                              "".format(self.container_name)

    @ensure_defined
    def load_config(self):
        self.config = {}
        with open(self.conf) as cf:
            # ignore comments
            for line in cf:
                if is_config_line(line):
                    k, v = line.split("=", 1)
                    self.config[k.strip()] = v.strip()

    @needs_root
    def create(self):
        if self.defined:
            raise InvalidConfigurationError("Duplicate container: {}"
                                            "".format(self.container_name))
        # we want to use the ubuntu template so that the environment is
        # consistent with what users expect, but we also make our own
        # adjustments to it as part of Project setup. It might be worth
        # building our own LXC template for Mojo at some point.
        dist = platform.dist()[0].lower()
        arch = ARCHITECTURES.get(platform.architecture()[0])
        if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
            template = 'ubuntu-cloud'
        else:
            template = dist
        cmd = ("lxc-create",
               "-t", template,
               "-n", self.container_name,
               )
        if os.path.isdir(os.path.dirname(self.container_root)):
            cmd += ("-B", "dir", "--dir", self.container_root)
        cmd += ("--",)
        if self.series is not None:
            cmd += ("--release", self.series)
        if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
            cmd += ("--arch", arch)
        logging.info("Creating LXC container '{}'".format(self.container_name))
        logging.info("Running command: {}".format(" ".join(cmd)))
        subprocess.check_call(cmd)
        logging.info("Setting up LXC")
        self.setup()
        logging.info("LXC creation Complete")

    @needs_root
    def _setup_legacy(self):
        pkgs_to_install = ['software-properties-common',
                           'python-software-properties',
                           'rsync']
        logging.info("Bootstrapping lxc env")
        self.load_config()
        # Serge Hallyn's solution to bootstrapping uses custom init scripts
        # The first init script runs whenever the lxc is started and it updates
        # and installs lxc and other packages needed
        with open(os.path.join(self.rootfs, 'etc', 'init',
                  'containersetup.conf'),
                  'w') as containersetup:
            containersetup.write(
                '''description "container setup job"
author "Serge Hallyn <serge.hallyn@canonical.com>"

start on runlevel [2345]
script
   apt-get update
   apt-get -y dist-upgrade
   apt-get -y install {}
end script\n'''.format(" ".join(pkgs_to_install)))
        # The second init script runs after the first has completed and
        # stops the lxc signaling completion of the installs
        # It also stops itself from running again by echoing 'manual'
        # into itself
        with open(os.path.join(self.rootfs, 'etc', 'init',
                  'setupstopped.conf'), 'w') as setupstopped:
            setupstopped.write(
                '''description "stop lxc containersetup has stopped"
author "Serge Hallyn <serge.hallyn@canonical.com>"
start on stopped containersetup

script
echo manual >> /etc/init/setup-stopped.conf
/sbin/poweroff
end script\n''')
        # Starting the lxc will run the init scripts
        self.start()
        # Wait for the lxc to stop signaling the completion of the installs
        self.wait("STOPPED")
        logging.info("Finished Bootstrapping")

        logging.info("Creating network-disabled and nested lxc configurations")
        shutil.copy(self.conf, self.conf_virtnet)
        with open(self.conf_virtnet) as virtnet:
            with open(self.conf_nonet, "w") as nonet, open(self.conf,
                                                           "w") as hostnet:
                for line in virtnet:
                    # Handle cgroups in a nested lxc per Serge Hallyn
                    line = line.replace(
                        "#lxc.aa_profile = lxc-container-default-with-nesting",
                        "lxc.aa_profile = lxc-container-default-with-nesting")
                    if not line.strip().startswith("lxc.network."):
                        hostnet.write(line)
                        nonet.write(line)
                nonet.write("lxc.network.type=empty\n")
        self.load_config()
        logging.info("LXC Setup Complete")

    @needs_root
    def _setup(self):
        logging.info("Creating network-disabled and nested lxc configurations")
        shutil.copy(self.conf, self.conf_nonet)
        with open(self.conf_nonet, "w") as nonet:
            with open(self.conf, "r") as hostnet:
                for line in hostnet:
                    # Handle cgroups in a nested lxc per Serge Hallyn
                    line = line.replace(
                        "#lxc.aa_profile = lxc-container-default-with-nesting",
                        "lxc.aa_profile = lxc-container-default-with-nesting")
                    if not line.strip().startswith("lxc.network."):
                        nonet.write(line)
                nonet.write("lxc.network.type=empty\n")
        self.load_config()
        logging.info("LXC Setup Complete")

    @needs_root
    def setup(self):
        if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
            self._setup()
        else:
            self._setup_legacy()

    def exists(self):
        return os.path.exists(self.conf)

    def is_setup(self):
        if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
            sentinel_file = self.conf_nonet
            # XXX: LP#1697574: Need to handle when conf_nonet doesn't exist for
            # previous set ups.
            if not os.path.exists(sentinel_file):
                logging.warn("Project doesn't seem properly set up, please fix "
                             "as project will break in future releases. {}"
                             .format(sentinel_file))
                return True
        else:
            sentinel_file = self.container_root + "/etc/init/containersetup.conf"
        return os.path.exists(sentinel_file)

    @needs_root
    @ensure_defined
    def destroy(self, force=False):
        cmd = ("lxc-destroy", "-n", self.container_name)
        if force:
            cmd += ("-f",)
        subprocess.check_output(cmd)

    def check_state(self, state=None):
        if state:
            cmd = ('sudo', 'lxc-info', '-s', '-n', self.container_name)
            logging.debug("Running command: {}".format(" ".join(cmd)))
            containers = yaml.safe_load(subprocess.check_output(cmd))
            # lxc-info shows 'state' (precise) or 'State' (newer)
            return state in containers.get('state', containers.get('State', {}))
        cmd = ('sudo', 'lxc-ls', '-1')
        containers = subprocess.check_output(cmd).split()
        return self.container_name in containers

    @ensure_defined
    def start(self, network=True):
        if self.running:
            logging.info("LXC container '{}' is already running"
                         "".format(self.container_name))
        else:
            logging.info("Starting LXC container '{}'".format(
                self.container_name))
            # always start containers in the background
            cmd = ["sudo", "lxc-start", "-d", "-n", self.container_name]
            keys = {}
            if not network:
                cmd += ["-f", self.conf_nonet]
            cmd += ["-s {}={}".format(k, v) for k, v in keys.items()]
            running = False
            logging.debug("Running command: {}".format(" ".join(cmd)))
            subprocess.check_call(cmd)
            lxc_retries = 0
            max_lxc_retries = 3
            running = self.wait("RUNNING", network)
            while running is False:
                lxc_retries += 1
                if lxc_retries > max_lxc_retries:
                    logging.debug("Unable to start LXC properly after {} attempts, giving up".format(max_lxc_retries))
                    break
                logging.debug("Possible LXC bug detected <https://github.com/lxc/lxc/issues/1237>, will try to work around it")
                self.stop()
                logging.debug("Re-running command (attempt {}/{}): {} ".format(lxc_retries, max_lxc_retries, " ".join(cmd)))
                subprocess.check_call(cmd)
                running = self.wait("RUNNING", network)

    @ensure_defined
    def stop(self):
        if not self.running:
            logging.info("LXC container '{}' is already stopped"
                         "".format(self.container_name))
        else:
            logging.info("Stopping LXC to container '{}'".format(
                self.container_name))
            cmd = ("sudo", "lxc-stop", "-n", self.container_name)
            subprocess.check_call(cmd)

    @ensure_defined
    def wait(self, state='RUNNING', network=False):
        cmd = ("sudo", "lxc-wait", "-n", self.container_name, "-s", state)
        logging.info("Waiting for LXC container to get to state '{}'"
                     "".format(state))
        subprocess.check_call(cmd)
        if network:
            networking = False
            attempts = 0
            max_attempts = 30
            interval = 1
            while networking is False:
                cmd = ('sudo', 'lxc-info', '-i', '-n', self.container_name)
                logging.debug("Running command: {}".format(" ".join(cmd)))

                # lxc-info -i will return a row for each IP, plus a newline
                found_ips = subprocess.check_output(cmd).split('\n')
                if len(found_ips) > 1:
                    networking = True
                else:
                    logging.debug("No networking in the LXC yet - retrying")

                attempts += 1
                time.sleep(interval)

                if attempts >= max_attempts:
                    logging.error("No networking after {} attempts, giving up".format(attempts))
                    cmd = ('sudo', 'lxc-info', '-n', self.container_name)
                    lxc_info_output = subprocess.check_output(cmd).split('\n')
                    logging.error("lxc-info output: {}".format(lxc_info_output))
                    return False
            return True
        return True

    @ensure_defined
    def run(self, command, env=None, user=None, network=True, stdout=None,
            stderr=None):
        "Run commands in the LXC environment"
        if not command:
            logging.warn("Refusing to execute empty command")
            return -1
        if not user:
            if os.environ.get('SUDO_USER'):
                user = os.environ.get('SUDO_USER')
            else:
                user = pwd.getpwuid(os.getuid()).pw_name
        # Are we attached to a tty? If not we'll need to create a logfile
        # for later use
        tty_attached = sys.stdout.isatty()
        if not tty_attached:
            lxc_console_log = tempfile.NamedTemporaryFile(delete=False)
            lxc_console_log.close()
        # Avoid looping fork bomb by not running when called as user
        # root
        if user != 'root':
            if not self.utils.user_exists(user):
                self.utils.create_user(user)
        command = ["sudo", "-E", "-u", user] + shlex.split(command)
        if isinstance(env, dict):
            env_vars = ["env"] + ["=".join((k, v)) for k, v in env.items()]
            command = env_vars + command
        if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
            self.stop()
            self.start(network=network)
            cmd = ["sudo", "lxc-attach", "-n", self.container_name]
        else:
            cmd = ["sudo", "lxc-start", "-n", self.container_name, "-F"]
            if not tty_attached:
                cmd.extend(["-L", lxc_console_log.name])
            if network:
                cmd.extend(["--share-net", "1"])
        cmd.extend(["--"])
        cmd.extend(command)
        logging.info("Running command in container '{}': {}"
                     "".format(self.container_name, ' '.join(command)))
        try:
            output = subprocess.check_output(cmd, stderr=stderr)
            if output.strip():
                logging.info(output)
        finally:
            # Output not yet displayed if we don't have a tty
            if not tty_attached:
                try:
                    with open(lxc_console_log.name, 'r') as clog:
                        logging.info(clog.read())
                    os.remove(lxc_console_log.name)
                except OSError:
                    logging.error("Unable to read or delete output of commands run inside LXC")
            if UBUNTU_RELEASES.get(self.series) >= UBUNTU_RELEASES['vivid']:
                self.stop()
        # subprocess.check_output will raise an error if it has a non-zero
        # return code in any case, so we can just return 0 here.
        return 0

    @property
    def defined(self):
        return self.check_state()

    @property
    def running(self):
        return self.check_state("RUNNING")

    @property
    def container_config_path(self):
        return os.path.join('/var/lib/lxc', self.container_name)

    @property
    @ensure_defined
    def rootfs(self):
        return self.config['lxc.rootfs']

    @property
    def conf(self):
        return os.path.join(self.container_config_path, "config")

    @property
    def conf_virtnet(self):
        return os.path.join(self.container_config_path, "config.virtnet")

    @property
    def conf_nonet(self):
        return os.path.join(self.container_config_path, "config.nonet")

    @property
    def fstab(self):
        return os.path.join(self.container_config_path, "fstab")


class ContainerUtils(object):
    def __init__(self, container):
        super(ContainerUtils, self).__init__()
        self.container = container

    def add_repo_key(self, repo_key):
        command = 'apt-key add {}'.format(repo_key)
        self.container.run(command, user="root")

    def add_repo(self, repo_spec, confirm=True):
        if confirm:
            confirmation = "-y"
        else:
            confirmation = ""
        command = 'add-apt-repository {} "{}"'.format(confirmation, repo_spec)
        self.container.run(command, user="root")
        self.apt_update()

    def apt_update(self):
        with open(os.devnull, 'w') as devnull:
            self.container.run("env DEBIAN_FRONTEND=noninteractive "
                               "apt-get update", user="root", stdout=devnull)
            self.container.run("env DEBIAN_FRONTEND=noninteractive "
                               "apt-get -y dist-upgrade", user="root",
                               stdout=devnull)

    def install_packages(self, *packages, **kwargs):
        prompt = kwargs.get("prompt", False)
        if prompt:
            assume_yes = ""
        else:
            assume_yes = "-y"
        command = "apt-get --no-install-recommends {} install {}".format(assume_yes,
                                                                         " ".join(packages))
        self.container.run(command, user="root")

    def user_exists(self, user):
        command = "getent passwd {}".format(user)
        try:
            with open(os.devnull, 'w') as devnull:
                self.container.run(command, user="root", stdout=devnull)
            return True
        except Exception:
            return False

    def create_user(self, user):
        (username, shadow, uid,
         gid, name, home, shell) = subprocess.check_output(
            ['getent', 'passwd', user]).split(':')
        # if the uid exists update the name to match the caller
        try:
            command = "useradd --home {} --create-home --uid {} {}" \
                "".format(home, uid, username)
            self.container.run(command, user="root")
        except subprocess.CalledProcessError as e:
            # LXC returns 4
            if e.returncode == 4 and self.container.container_class == 'lxc':
                command = "cp /etc/sudoers.d/ubuntu-all " \
                    "/etc/sudoers.d/{}-all".format(username)
                self.container.run(command, user="root")
                command = 'sed -i "s/ubuntu/{}/g" ' \
                    '/etc/sudoers.d/{}-all'.format(username, username)
                self.container.run(command, user="root")
                command = 'usermod -l {} -G sudo ubuntu'.format(username)
                self.container.run(command, user="root")
            # LXD can return 4 or 255
            elif e.returncode in [4, 255] and self.container.container_class == 'lxd':
                with tempfile.NamedTemporaryFile() as temp:
                    temp.write('{} ALL=(ALL) NOPASSWD:ALL'.format(username))
                    temp.flush()
                    self.container.put(temp.name, '/etc/sudoers.d/{}-all'.format(username), mode=0o400)
                command = 'usermod -l {} -G sudo ubuntu'.format(username)
                self.container.run(command, user="root")
            else:
                raise e


class NoContainer(Container):
    def __init__(self, name, series, ensure_exists=False):
        Container.__init__(self, name, series, ensure_exists)
        self.container_class = 'containerless'
        self.container_root = ''

    def create(self):
        logging.info("Using no container. Not creating a container")

    def destroy(self, force=False):
        logging.info("Using no container. Not destroying a container")

    def run(self, command, env=None, user=None, network=True, stdout=None,
            stderr=None):
        if isinstance(env, dict):
            env_vars = ["env"] + ["=".join((k, v)) for k, v in env.items()]
            command = env_vars + shlex.split(command)
        logging.warn("Atempting to run in a container with a containerless "
                     "project. Running command without a container: {}"
                     "".format(command))
        output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        if output.strip():
            logging.info(output)

    def exists(self):
        return os.path.exists(self.container_root)

    def is_setup(self):
        return self.exists()

    @property
    def defined(self):
        return True

    @property
    def conf(self):
        logging.info("Containerless project has no container config")
        return self.container_root
