

class JujuCheckError(Exception):
    pass


class JujuCheck(object):
    cmd_name = None  # A simple command name used to call this check when used with juju-check
    help = None  # Help string for this check

    def __init__(self, status):
        """
        Args:
            status: A mojo.juju.status.JujuStatusBase object
        """
        self.status = status

    def check(self, args):
        """ Run the check raising an exception on problems, returning a string of output otherwise.
        """
        raise NotImplementedError

    @staticmethod
    def setup_argparse(ap):
        """ Adds arguent options to an argument parser. If no arguments are needed this method can be skipped.
        Args:
            ap: An argumentparser
        """
        return


class Errors(JujuCheck):
    cmd_name = 'errors'
    help = "Checks the status for any errors, raising an exception if found."

    def check(self, args):
        self.status.ready()
        return "No Juju status errors found"


class Unused(JujuCheck):
    cmd_name = 'unused_machines'
    help = "Checks for any machines that are allocated but unused by any applications."

    def check(self, args):
        used_machines = set()
        for app in self.status.applications_list():
            used_machines.update(self.status.application_machine_numbers(app))

        all_machines = set(self.status.yaml_status()['machines'].keys())

        unused_machines = all_machines - used_machines
        if len(unused_machines) > 0:
            raise JujuCheckError("Unused machines found: {}".format(list(unused_machines)))

        return "All machines are being used"


class UpgradeAvailable(JujuCheck):
    cmd_name = 'upgrade_available'
    help = "Checks whether the controller has a newer Juju version than the model is running"

    def check(self, args):
        if args.check_version:
            controller_version = args.check_version
        elif args.check_version_file:
            with open(args.check_version_file, 'r') as f:
                controller_version = f.readline().split()[0]
        else:
            controller_version = self.status.controller_version()

        # I assume the model can't have a newer version than the controller
        if controller_version != self.status.model_version():
            raise JujuCheckError("controller version {} != model version {}".format(controller_version,
                                                                                    self.status.model_version()))

        return "The Juju controller and model are running the same version ({})".format(controller_version)

    @staticmethod
    def setup_argparse(ap):
        ap.add_argument('--check_version', help='A controller version to check against by the default the' +
                                                ' controller version is determined automatically.')
        ap.add_argument('--check_version_file', help='A file containing the controller version to check against by ' +
                                                     'the default the controller version is determined automatically.')
