import ast
import logging
import subprocess
import yaml

import mojo.juju

from mojo.exceptions import ConfigNotFoundException

DEFAULT_DEBUG_LOG_CONFIG = {}


class UnsupportedDebugLogType(Exception):
    pass


class DebugLogs(object):
    def __init__(self, workspace, stage, debug_logs, debug_logs_stages_to_exclude):
        self.debug_logs = debug_logs or 'debug-logs'
        self.debug_logs_stages_to_exclude = debug_logs_stages_to_exclude or 'production'
        self.stage = stage
        self.stage_name = stage.split('/')[-1]
        self.workspace = workspace
        self.supported_types = {
            'config-files': {
                'description': 'Configuration Files',
                'command': 'cat'},
            'directory-listings': {
                'description': 'Directory Listing of',
                'command': 'ls -lha'},
            'logs': {
                'description': 'Log Files',
                'command': 'tail -n 200'},
            }
        self._config = None

    @property
    def has_config(self):
        if self.config:
            return True
        return False

    @property
    def config(self):
        if not self._config:
            self._config = self._load_config()
        return self._config

    def get_log_output(self):
        self._check_config()
        log_output = []
        if self.config:
            logging.debug("Debug Log Config found: {}".format(self.config))
            for log in self._gather_logs():
                log_output.append("###############################################")
                log_output.append("Debug Logs: {}".format(log['description'].strip()))
                log_output.append("###############################################")
                log_output.append(log['content'].strip())
        else:
            log_output.append("No debug-logs to gather")
        return "\n".join(log_output)

    def _check_config(self):
        for service in self.config:
            for debug_log_type in self.config[service]:
                if debug_log_type not in self.supported_types:
                    raise UnsupportedDebugLogType(
                        "Unsupported debug-log type: {}. Supported types are: {}".format(
                            debug_log_type, ", ".join(self.supported_types)))

    def _load_config(self):
        if self.stage_name in self.debug_logs_stages_to_exclude:
            return

        try:
            with open(self.workspace.spec.get_config(self.debug_logs, self.stage), 'r') as debug_log_file:
                debug_config = yaml.safe_load(debug_log_file)
        except ConfigNotFoundException:
            # Just use default debug-log config
            logging.warning("No debug log matching {} found. Using default.".format(self.debug_logs))
            debug_config = DEFAULT_DEBUG_LOG_CONFIG
        return debug_config

    def _resolve_remote_paths(self, unit, path):
        cmd = ['juju', 'run', '--unit', unit,
               'PYTHON=$(which python3 python | head -n1) ; '
               '[ -z "$PYTHON" ] && {{ echo ERROR: No python3 or python found. ; exit 1 ; }} ; '
               '$PYTHON -c "import glob; print(glob.glob(\'{}\'))"'.format(path)]
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        return ast.literal_eval(output)

    def _run_remote_command(self, unit, command, sanitized_path_list):
        cmd = ['juju', 'run', '--unit', unit, "{} {}".format(command, " ".join(sanitized_path_list))]
        return subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    def _gather_logs(self):
        output = []
        status = mojo.juju.Status()
        for service in self.config:
            for unit in status.service_units(service) or []:
                for debug_log_type in self.config[service]:
                    command = self.supported_types[debug_log_type]['command']
                    for path in self.config[service][debug_log_type]:
                        sanitized_path_list = []
                        try:
                            sanitized_path_list = self._resolve_remote_paths(unit, path)
                            if sanitized_path_list:
                                try:
                                    content = self._run_remote_command(unit, command, sanitized_path_list)
                                except subprocess.CalledProcessError as e:
                                    logging.error("There was an error running '{} {}' on {}".format(
                                        command, " ".join(sanitized_path_list), unit))
                                    logging.error("Error was: {}".format(e.output))
                                    content = "Unable to run '{} {}' on {}".format(
                                        command, " ".join(sanitized_path_list), unit)
                            else:
                                content = "Path {} doesn't exist on {} for the {} command".format(path, unit, debug_log_type)
                        except subprocess.CalledProcessError as e:
                            logging.error("There was an error resolving path {} for unit {} as part of {}".format(
                                path, unit, debug_log_type))
                            logging.error("Error was: {}".format(e.output))
                            content = "Unable to resolve path {} for unit {} as part of {}".format(path, unit, debug_log_type)
                        except SyntaxError as e:
                            logging.error("There was an error resolving path {} for unit {} as part of {}".format(
                                path, unit, debug_log_type))
                            logging.error("Error was: {}".format(e))
                            logging.error("Unable to parse: {}".format(e.text))
                            content = "Unable to resolve path {} for unit {} as part of {}".format(path, unit, debug_log_type)
                        if content:
                            output.append({
                                "description": "{} {} on {}".format(
                                    self.supported_types[debug_log_type]['description'],
                                    ", ".join(sanitized_path_list),
                                    unit),
                                "content": content})
        return output
