""" Provides convenience commands for extracting information from Juju status
    Juju version 2 only.
"""
import argparse
import sys

from status import Juju2Status


class StatusInfo:
    name = None  # A simple command name used to call this
    help = None  # Help string for this check

    def __init__(self, status):
        """
        Args:
            status: A mojo.juju.status.JujuStatusBase object
        """
        self.status = status

    def run(self, args):
        """ Run returning a string of output.
        """
        raise NotImplementedError

    @staticmethod
    def setup_argparse(ap):
        """ Adds arguent options to an argument parser. If no arguments are needed this method can be skipped.
        Args:
            ap: An argumentparser
        """
        return


class GetIPs(StatusInfo):
    name = 'get_ips'
    help = "Returns a list of ips. Specify either a space separated list of application names or 'all'."

    def run(self, args):
        ips = []
        if 'all' in args.applications:
            for machine in self.status.yaml_status()['machines'].values():
                if 'dns-name' in machine:
                    ips.append(machine['dns-name'])
                if 'containers' in machine:
                    for container in machine['containers'].values():
                        if 'dns-name' in container:
                            ips.append(container['dns-name'])
        else:
            applications = args.applications
            for app in applications:
                ips.extend(self.status.application_ips(app))

        return "\n".join(ips)

    @staticmethod
    def setup_argparse(ap):
        ap.add_argument('applications', nargs='+')


def main():

    parser = argparse.ArgumentParser(description="Provides convenience commands for parsing Juju2 status.")
    parser.add_argument("-f", "--file", dest="status_file",
                        help="A yaml status file; if unspecified, status is pulled from the environment.")

    parser.add_argument("-t", "--command-timeout", type=int, default=300,
                        help="Kill 'juju status' and give up if this amount of time elapses.")

    sp = parser.add_subparsers(title="", dest="status_info")

    # Setup all available StatusInfo classes as subparsers
    available_info = {'get_ips': GetIPs}
    for cmd in available_info.values():
        sub = sp.add_parser(cmd.name, description=cmd.help)
        cmd.setup_argparse(sub)

    args = parser.parse_args()

    if args.status_file is not None:
        with open(args.status_file, 'r') as sf:
            status_yaml = sf.read()
        status = Juju2Status(status_yaml=status_yaml, command_timeout=args.command_timeout)
    else:
        status = Juju2Status(command_timeout=args.command_timeout)

    print(available_info[args.status_info](status).run(args))
    return 0


if __name__ == "__main__":
    sys.exit(main())
