import logging
import shlex

from mojo import phase
from mojo.exceptions import InvalidManifestException


class Manifest(object):
    def __init__(self, spec, manifest_file, stage):
        super(Manifest, self).__init__()
        self.spec = spec
        self.stage = stage
        self.config = spec.get_config(base=manifest_file, stage=stage)
        self.phases = []
        self.load()

    def load(self):
        self.phases = self.parse_manifest(self.config)

    def parse_comment(self, bareline):
        text_comment = bareline[1:].strip()
        return "comment", dict(text=text_comment)

    def parse_command(self, bareline):
        words = shlex.split(bareline)
        command = words[0]
        kwargs = {}
        for arg in words[1:]:
            try:
                k, v = arg.split("=", 1)
            except ValueError:
                raise InvalidManifestException("Not a manifest line: '{}'".format(bareline))

            if k in kwargs:
                raise InvalidManifestException(
                    '{} option {}={} would override '
                    'previously supplied option {}={}.'.format(
                        command, k, v, k, kwargs[k]))

            kwargs[k] = v
        return command, kwargs

    def parse_manifest(self, manifest_path):
        """Return a list of phases from the manifest file."""
        phases = []
        with open(manifest_path) as manifest_file:
            previous_phase = None
            for line in manifest_file:
                bareline = line.strip()
                if not bareline or bareline.startswith("##"):
                    continue

                if bareline.startswith("#"):
                    command, kwargs = self.parse_comment(bareline)
                else:
                    command, kwargs = self.parse_command(bareline)

                config = kwargs.get('config')
                if command == "include" and config:
                    try:
                        phases.extend(self.parse_manifest(self.spec.get_config(base=config, stage=self.stage)))
                    except InvalidManifestException as e:
                        raise InvalidManifestException(
                            "Unable to parse include: '{}' {}".format(bareline, e))
                    previous_phase = None  # prevent comment-glomming
                    continue

                try:
                    current_phase = phase.build(command, **kwargs)
                except phase.NoSuchPhaseError as e:
                    raise InvalidManifestException(e.message)

                if (previous_phase and previous_phase.name == "comment" and
                        current_phase.name == "comment"):
                    previous_phase.options['text'] += (
                        "\n" + current_phase.options['text'])
                else:
                    phases.append(current_phase)

                previous_phase = current_phase
        return phases

    def run(self, project, workspace, stage=None, interactive=False):
        "Run all phases defined in this manifest"
        auto_repo = True
        if "repo" in [p.name for p in self.phases]:
            # Repo step is being called explicitly - don't call it
            # automatically
            auto_repo = False
        auto_secrets = True
        if "secrets" in [p.name for p in self.phases]:
            # Secrets are being called explicitly - don't call it
            # automatically
            auto_secrets = False
        for phase_name in self.phases:
            run_phase = True
            if phase_name.name != "comment" and interactive:
                logging.info("Interactive mode... about to run:")
                logging.info("  {} {}".format(
                    phase_name.name,
                    " ".join(['{}={}'.format(i, phase_name.options[i]) for i in phase_name.options.keys()])))
                logging.info("")
                logging.info("[(Y)es] / (N)o / (S)top")
                while True:
                    action = raw_input("Run this action? ([y]/n/s) ")
                    if not action or action[0].lower() == "y":
                        # run_phase is True, so just carry on.
                        break
                    elif action[0].lower() == "s":
                        return
                    elif action[0].lower() == "n":
                        logging.info("Skipping...")
                        logging.info("")
                        run_phase = False
                        break
                    logging.info("Didn't recognize option {}, please try again".format(action))
            if not run_phase:
                continue
            if phase_name.name in ["deploy", "build", "script", "verify"]:
                phase_name.run(project, workspace, stage,
                               auto_secrets=auto_secrets)
            elif phase_name.name == "collect":
                phase_name.run(project, workspace, stage,
                               auto_repo=auto_repo)
            else:
                phase_name.run(project, workspace, stage)
