import errno
import json
import logging
import os
import pwd
import re
import shutil
import sys

from mojo.contain import (
    LXCContainer,
    LXDContainer,
    NoContainer,
    NoSuchContainerError,
    )
from mojo.exceptions import ArgumentError
from mojo.utils import (
    mkdirs,
    mksymlink
    )
from mojo.workspace import Workspace


RESERVED_PROJECT_NAMES = [
    'LOCAL',
    'bin',
    'lib',
    ]


class DuplicateProjectError(Exception):
    pass


class ProjectCreationError(Exception):
    pass


class ProjectNameError(Exception):
    pass


class ProjectDoesNotExistError(Exception):
    pass


class Project(object):
    "A mojo project"
    token_re = re.compile(r"^[a-zA-Z0-9_.-]+$").search
    token_desc = "{} must only contain alphanumerics, underscores, or periods."

    def __init__(self, name, series, mojo_root=None, container_class=LXCContainer):
        if not name:
            raise ProjectCreationError("No project name specified. Use "
                                       "--project or set MOJO_PROJECT")
        if not bool(self.token_re(name)):
            raise ProjectNameError(self.token_desc.format("Project names"))
        if name in RESERVED_PROJECT_NAMES:
            raise ProjectNameError("Projects may not be named {}".format(name))
        self.mojo_roots = []
        if mojo_root:
            self.mojo_roots.append(mojo_root)
        if container_class == LXCContainer:
            # Privileged location first, then unprivileged
            self.mojo_roots.append('/srv/mojo')
            self.mojo_roots.append(os.path.expanduser('~/.local/share/mojo'))
        else:
            # Unprivileged first, then privileged
            self.mojo_roots.append(os.path.expanduser('~/.local/share/mojo'))
            self.mojo_roots.append('/srv/mojo')
        self.name = name
        if series is None:
            raise ArgumentError("Series setting missing. Use --series or set MOJO_SERIES")
        self.series = series

        # map my running uid to username
        self.owner = pwd.getpwuid(os.getuid()).pw_name
        # If we're root, we may be running on behalf of $SUDO_USER,
        # and therefore any files we create should belong to them.
        if os.getuid() == 0:
            self.owner = os.environ.get('SUDO_USER', self.owner)

        self.project_info = self.get_project_info()

        # ### Paths ###
        # <mojo_root>/<prj>
        if self.project_info.get('project_root'):
            self.root = self.project_info.get('project_root')
        else:
            self.root = Project.root_path(self.mojo_roots[0], self.name)
        self.mojo_root = os.path.normpath(os.path.join(self.root, '..'))

        # <mojo_root>/<prj>/<series>
        self.project_series = os.path.join(self.root, self.series)

        if self.project_info.get('container_class'):
            logging.info("Using pre-existing container class {}"
                         "".format(self.project_info.get('container_class')))
            if self.project_info.get('container_class') == 'lxc':
                container_class = LXCContainer
            elif self.project_info.get('container_class') == 'lxd':
                container_class = LXDContainer
            elif self.project_info.get('container_class') == 'containerless':
                container_class = NoContainer
            # else use container_class passed to init
            # this defaults to container_class=LXCContainer
        self.container = container_class(self.name, self.series)
        if self.container.container_class == 'lxc':
            # <mojo_root>/<prj>/<series>/ROOTFS
            self.project_container_root = os.path.join(self.project_series, "ROOTFS")
            # <mojo_root>/<prj>/<series>/ROOTFS/<mojo_root>/<prj>/<series>
            self.container_project_series = self.project_container_root + self.project_series
            if self.project_info.get('container_root'):
                self.container.container_root = self.project_info.get('container_root')
            # If the lxc does not already exist use location setting
            elif not self.container.exists():
                self.container.container_root = self.project_container_root
        elif self.container.container_class in ['containerless', 'lxd']:
            self.container.container_root = self.project_series
            self.project_container_root = self.project_series
            self.container_project_series = self.project_series

    def get_project_info(self):
        """Get pre-existing project information from the .project file.
        Return a dictionary of pre-existing information."""
        for mojo_root in self.mojo_roots:
            info_file = os.path.join(mojo_root, self.name, self.series, '.project')
            if os.path.exists(info_file):
                with open(info_file, 'r') as info:
                    return json.loads(info.read())
        return {}

    def write_project_info(self):
        """Create a project information file. This will allow for state
        to be learned for subsequent mojo commands without requiring prior
        knowledge of the pre-existing project or command line switches"""
        info_file = os.path.join(self.project_series, '.project')
        # This can be expanded down the road with more data
        project_info = {'container_class': self.container.container_class,
                        'container_root': self.container.container_root,
                        'project_root': self.root}
        with open(info_file, 'w') as info:
            info.write(json.dumps(project_info))

    @classmethod
    def root_path(cls, mojo_root, name):
        if not bool(cls.token_re(name)):
            raise ProjectNameError(cls.token_desc.format("Project names"))
        if name in RESERVED_PROJECT_NAMES:
            raise ProjectNameError("Projects may not be named {}".format(name))

        return os.path.join(mojo_root, name)

    @classmethod
    def list_series(cls, mojo_root, name):
        if not bool(cls.token_re(name)):
            raise ProjectNameError(cls.token_desc.format("Project names"))
        if name in RESERVED_PROJECT_NAMES:
            raise ProjectNameError("Projects may not be named {}".format(name))

        try:
            return [s for s in sorted(os.listdir(cls.root_path(mojo_root, name)))]
        except OSError as ose:
            if ose.errno == errno.ENOENT:
                raise ProjectDoesNotExistError(
                    'No such project: {}'.format(name))
            raise

    @classmethod
    def list(cls, mojo_root, only_names):
        """List all Mojo projects"""
        mojo_roots = ['/srv/mojo', os.path.expanduser('~/.local/share/mojo')]
        if mojo_root:
            mojo_roots.append(mojo_root)
        proj_list = []
        for mojo_root in set(mojo_roots):
            if os.path.isdir(mojo_root):
                for proj in os.listdir(mojo_root):
                    # Ignore any directories we don't have permissions to read in case
                    # we have a multi-tenant mojo install. Now that we support setting mojo_root
                    # that should be the solution for new mojo installs but this is still needed
                    # for backward compatibility
                    try:
                        os.listdir(os.path.join(mojo_root, proj))
                    except OSError:
                        continue
                    if proj in RESERVED_PROJECT_NAMES:
                        continue
                    if not only_names:
                        series = " ".join(cls.list_series(mojo_root, proj))
                        proj_list.append("{}: {}".format(proj, series))
                    else:
                        proj_list.append(proj)
        return proj_list

    def create(self):
        "Create a new container and project"
        if self.container.container_class == "lxc":
            if os.geteuid() != 0:
                raise ProjectCreationError(
                    "Could not create lxc without running as a privileged user")
        logging.info("Checking {} project path at {} ...".format(
            self.name, self.root))
        if os.path.exists(self.root):
            logging.warning("Project {} already exists at {}. Using existing "
                            "project".format(self.name, self.root))
        else:
            try:
                mkdirs(self.root, self.owner, mode=0700)
            except OSError as e:
                raise ProjectCreationError(
                    "Could not create project root: {}".format(e.strerror))

        if os.path.exists(self.project_series):
            logging.warning("Project {} project_series already exists at {}, "
                            "using it.".format(self.name, self.project_series))
        else:
            try:
                mkdirs(self.project_series, self.owner, mode=0700)
            except OSError as e:
                raise ProjectCreationError(
                    "Could not create project project_series: "
                    "{}".format(e.strerror))

        logging.info("Checking {} project container at {} ..."
                     "".format(self.name, self.container.conf))
        if not self.container.exists():
            logging.info("Not found, creating container")
            self.container.create()

        logging.info("Checking {} project container is setup ..."
                     "".format(self.name))
        if not self.container.is_setup():
            logging.info("Not setup - setting up container for mojo")
            self.container.setup()

        # Only needed for lxc. We mount the filesystem in container.setup() above in lxd and commands are run via
        # the LXD API so no need for sudo perms.
        if self.container.container_class == 'lxc':
            if self.project_container_root != self.container.container_root:
                logging.info("Checking {} project container symlink {} -> {} ..."
                             "".format(self.name, self.container.container_root, self.project_container_root))
                if not os.path.exists(self.project_container_root):
                    os.symlink(self.container.container_root, self.project_container_root)

            logging.info("Checking {} project dir {} ..."
                         "".format(self.name, self.container_project_series))
            if not os.path.exists(self.container_project_series):
                mkdirs(self.container_project_series, self.owner, mode=0700)

            sudoers_file = os.path.join(self.project_container_root, "etc", "sudoers.d",
                                        "ubuntu-all")
            logging.info("Writing {} sudoers file: {}".format(
                         self.name, sudoers_file))
            with open(sudoers_file, "w") as f:
                f.write("ubuntu ALL=(ALL) NOPASSWD: ALL\n")
            os.chmod(sudoers_file, 0400)

        if not os.path.exists(os.path.join(self.project_series, '.project')):
            self.write_project_info()

    def delete(self, force=True):
        """Delete a project and its container - if no series remain, also delete project root"""
        try:
            self.container.destroy()
        except NoSuchContainerError:
            if not force:
                raise
        try:
            shutil.rmtree(self.project_series)
            logging.debug("Deleted {}".format(self.project_series))
        except OSError:
            if not force:
                raise

        # If no series remain, delete project root
        if not Project.list_series(self.mojo_root, self.name):
            rpath = Project.root_path(self.mojo_root, self.name)
            logging.info("Destroying project root: {}"
                         "".format(rpath))
            try:
                shutil.rmtree(rpath)
            except OSError:
                if force:
                    logging.warning("Unable to remove project root '{}'. Please remove manually".format(rpath))
                else:
                    raise

    def ws_link(self, title):
        return os.path.join(self.project_series, str(title))

    def create_workspace(self, spec_url, title, stage=None,
                         manifest_file='manifest', logfile=None):

        # Validate the project exists
        if self.name not in self.list(self.mojo_root, True):
            print "Project, '{}', does not appear to exist.".format(self.name)
            print "Please run 'mojo project-new --series {} {}' to " \
                  "create the project.".format(self.series, self.name)
            print "Or check the file permissions on {}".format(self.root)
            sys.exit(1)
        if not self.container.exists() or not self.container.is_setup():
            print "The container project and series path for {} does not appear " \
                  "to exist at {}.".format(self.name, self.container_project_series)
            print "Check to make sure the project and series exist and " \
                  "permissions are set properly."
            print "Please run 'mojo project-new --series {} {}' to " \
                  "create the project if needed.".format(self.series, self.name)
            print "Or check the file permissions on {}".format(self.container_project_series)
            sys.exit(1)

        # If we're using nested directories for specs, make sure our workspace
        # symlink doesn't need additional sub-directories created.
        if "/" in title:
            title = title.replace("/", "-")
        ws_path = os.path.join(self.container_project_series, str(title))
        mkdirs(ws_path, self.owner)
        if self.container.container_class in ['containerless', 'lxd']:
            mkdirs(self.ws_link(title), self.owner)
        else:
            mksymlink(ws_path, self.ws_link(title))
        ws = Workspace(title, self.ws_link(title), stage=stage, spec_url=spec_url,
                       manifest_file=manifest_file, logfile=logfile)
        ws.prepare()
        return ws

    def workspace(self, workspace_name, logfile=None):
        return Workspace(workspace_name, self.ws_link(workspace_name),
                         logfile=logfile)

    def delete_workspace(self, workspace_name):
        ws = self.workspace(workspace_name)
        ws.delete()
        if os.path.lexists(self.ws_link(workspace_name)):
            os.unlink(self.ws_link(workspace_name))

    def workspace_list(self):
        workspaces = [w for w in os.listdir(self.container_project_series)
                      if w not in ['LOCAL', '.project']]
        workspace_output = ["Workspaces:", "==================="]
        for workspace in workspaces:
            workspace_output.append("{}".format(workspace))
        return workspace_output

    @property
    def utils(self):
        return self.container.utils
