# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import mock
import io
import os
from unittest import TestCase
from shutil import rmtree
from tempfile import mkdtemp

import mojo
from mojo.cli import (
    create_arg_parser,
    run_with_args,
    MojoArgumentParser,
    )
from mojo.exceptions import ArgumentError
import mojo.project
from mojo.tests.utils import create_unique_bzr_branch


class ArgParserTestCase(TestCase):
    """The argument parser includes the expected args."""

    def test_arg_parser(self):
        parser = create_arg_parser()

        args = parser.parse_args(
            "workspace-new --project my-project -s trusty "
            "--stage devel "
            "/home/me/my/spec_url my-workspace".split())

        self.assertEqual(args.workspace, 'my-workspace')
        self.assertEqual(args.project, "my-project")
        self.assertEqual(args.series, "trusty")
        self.assertEqual(args.spec_url, "/home/me/my/spec_url")
        self.assertEqual(args.stage, "devel")

    def test_invalid_spec_from_dir(self):
        parser = create_arg_parser()
        args = parser.parse_args("-r /srv/mojo workspace-new --project my-project "
                                 "-s trusty --stage devel /srv/mojo/my/spec/in/bad/location "
                                 "my-workspace".split())

        with self.assertRaises(ArgumentError):
            mojo.cli.get_spec_url_and_workspace(args)

    def test_munging_argument_parser(self):
        p = MojoArgumentParser()
        p.add_argument('-a', '--foo-bar', '--ying_yang', default='bingo')
        h = io.BytesIO()
        p.print_help(h)

        # An underscore in options is converted to a hyphen.
        self.assertTrue('--ying-yang' in h.getvalue().decode('UTF-8'))
        self.assertEqual(p.parse_args(['-a', 'val']).foo_bar, 'val')
        self.assertEqual(p.parse_args(['--foo-bar', 'val']).foo_bar, 'val')
        self.assertEqual(p.parse_args(['--ying-yang', 'val']).foo_bar, 'val')

        # Hypen and underscore are interchangable.
        self.assertEqual(p.parse_args(['--ying_yang', 'val']).foo_bar, 'val')

        # Non-options are not mangled.
        p.add_argument('required_arg')
        self.assertEqual(p.parse_args(['a_value-here']).required_arg,
                         'a_value-here')

        # And -- argument escaping works unmangled.
        self.assertEqual(p.parse_args(['--', '--ying_yang']).required_arg,
                         '--ying_yang')


class WorkspaceNewTestCase(TestCase):

    def create_spec_dir(self):
        spec_dir = mkdtemp()
        self.addCleanup(rmtree, spec_dir)
        create_unique_bzr_branch(spec_dir, ('services.yml',))
        return spec_dir

    @mock.patch('mojo.contain.LXCContainer.is_setup', return_value=True)
    @mock.patch('mojo.contain.LXCContainer.exists', return_value=True)
    def test_workspace_new(self, exists, is_setup):
        mojo_root = mkdtemp()
        self.addCleanup(rmtree, mojo_root)
        spec_dir = self.create_spec_dir()
        parser = create_arg_parser()
        args = parser.parse_args([
            "-r",
            mojo_root,
            "workspace-new",
            "--project",
            "my-project",
            "-s",
            "trusty",
            spec_dir,
            "my-workspace",
            ])

        # Mock the LXC directory
        lxc_project_series = os.path.join(mojo_root, 'my-project', 'trusty',
                                          'ROOTFS', mojo_root[1:],
                                          'my-project', 'trusty')
        mojo.utils.mkdirs(lxc_project_series)
        setup_file = os.path.join(mojo_root, 'my-project', 'trusty', 'ROOTFS',
                                  'etc', 'init', 'setup-stopped.conf')
        mojo.utils.mkdirs(os.path.dirname(setup_file))
        open(setup_file, 'a').close()

        # Mock Project.list()
        with mock.patch.object(mojo.project.Project, 'list',
                               return_value='my-project'):
            run_with_args(args)

        expected_dirs = ['build', 'local', 'log', 'spec']
        workspace_dir = os.path.join(mojo_root, 'my-project', 'trusty',
                                     'my-workspace')
        for expected_dir in expected_dirs:
            expected_path = os.path.join(workspace_dir, expected_dir)
            self.assertTrue(os.path.exists(expected_path))

    @mock.patch('mojo.contain.LXCContainer.is_setup', return_value=True)
    @mock.patch('mojo.contain.LXCContainer.exists', return_value=True)
    @mock.patch('mojo.project.Project.list', return_value='my-project')
    def test_workspace_new_precise(self, project_list, exists, is_setup):
        mojo_root = mkdtemp()
        self.addCleanup(rmtree, mojo_root)
        spec_dir = self.create_spec_dir()
        parser = create_arg_parser()
        args = parser.parse_args([
            "-r",
            mojo_root,
            "workspace-new",
            "--project",
            "my-project",
            "-s",
            "precise",
            spec_dir,
            "my-workspace",
            ])

        # Mock the LXC directory
        lxc_project_series = os.path.join(mojo_root, 'my-project', 'precise',
                                          'ROOTFS', mojo_root[1:],
                                          'my-project', 'precise')
        mojo.utils.mkdirs(lxc_project_series)
        setup_file = os.path.join(mojo_root, 'my-project', 'precise', 'ROOTFS',
                                  'etc', 'init', 'setup-stopped.conf')
        mojo.utils.mkdirs(os.path.dirname(setup_file))
        open(setup_file, 'a').close()

        run_with_args(args)

        expected_dirs = ['build', 'local', 'log', 'spec']
        workspace_dir = os.path.join(mojo_root, 'my-project', 'precise',
                                     'my-workspace')
        for expected_dir in expected_dirs:
            expected_path = os.path.join(workspace_dir, expected_dir)
            self.assertTrue(os.path.exists(expected_path))
