# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import tempfile
from unittest import TestCase

import os
import pylxd
from pylxd.exceptions import LXDAPIException

from mojo.contain import ContainerTimeout, InvalidConfigurationError, LXDContainer, NoSuchContainerError


class LXDContainerTest(TestCase):

    def setUp(self):
        self.c = LXDContainer("mojo-unit-tests", "xenial")
        self.lxd = pylxd.Client()

    def tearDown(self):
        try:
            c = self.lxd.containers.get(self.c.container_name)
        except LXDAPIException:
            pass
        else:
            if c.status == "Running":
                c.stop(wait=True)
            c.delete(wait=True)

    def test_create(self):
        raised = False
        try:
            LXDContainer("a.name.with.periods", "xenial")
        except InvalidConfigurationError:
            raised = True
        self.assertTrue(raised)

        orig_series = self.c.series
        self.c.series = "doesnotexist"
        self.assertRaises(NoSuchContainerError, self.c.create)

        self.c.series = orig_series
        # This test requires a LXD image with the correct alias for mojo
        self.c.create()
        # containers.get raises and exception if not found
        self.lxd.containers.get(self.c.container_name)

    def test_exists(self):
        """ Tests exists(), defined() and container property.
        """
        self.assertIsNone(self.c.container)
        self.assertFalse(self.c.exists())
        self.assertFalse(self.c.defined)

        self.c.create()

        want = self.lxd.containers.get(self.c.container_name)
        self.assertEqual(self.c.container.config, want.config)
        self.assertTrue(self.c.exists())
        self.assertTrue(self.c.defined)

    def test_container_states(self):
        """ Tests _check_state(), running(), start(), stop() and wait()
        """
        self.c.create()
        self.c.start()

        self.c._wait()
        self.assertEqual(self.c._check_state(), "Running")
        self.assertTrue(self.c.running)

        self.c.stop()
        self.assertEqual(self.c._check_state(), "Stopped")
        self.assertFalse(self.c.running)
        self.c.timeout = 4
        self.assertRaises(ContainerTimeout, self.c._wait)

    def test_mounts(self):
        """ Test _ensure_mount, is_setup and destroy
        """
        self.c.create()

        self.assertFalse(self.c.is_setup())

        self.c.start()
        tmpdir = tempfile.mkdtemp(prefix="mojo-test-container")
        self.c.container_root = tmpdir

        self.c._setup_mount()
        self.assertTrue(self.c.is_setup())

        self.c.destroy()
        self.assertFalse(self.c.is_setup())
        os.removedirs(tmpdir)

    def test_network(self):
        """ Test _enable_network, _disable_network and _is_network_enabled
        """
        self.c.create()

        self.assertTrue(self.c._is_network_enabled())

        self.c._disable_network()
        self.assertFalse(self.c._is_network_enabled())

        self.c._enable_network()
        self.assertTrue(self.c._is_network_enabled())

    def test_in_container_ops(self):
        """ Tests run() and put()
        """
        test_file = "/atestfile"
        self.c.create()
        devnull = open('/dev/null', 'w')

        self.assertEqual(1, self.c.run("ls " + test_file, stdout=devnull, stderr=devnull))

        self.c.put('/etc/services', test_file)
        self.assertEqual(0, self.c.run("ls " + test_file, stdout=devnull, stderr=devnull))
