# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import mock
import os
from StringIO import StringIO
import sys
from unittest import TestCase

import mojo.juju
from mojo.juju.check import main as check
from mojo.juju.parse_status import main as parse_status


class Juju2StatusTestCase(TestCase):

    @staticmethod
    def _testdata_from_file(filename):
        """ Loads testdata files.
        Args:
            filename: filename containing the desired data relative to the testdata dir.
        Returns: string of testdata.
        """
        base_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(base_dir, 'testdata', filename), 'r') as test_file:
            return test_file.read()

    def setUp(self):
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        self.status._raw_status = self._testdata_from_file('juju2-idle.yaml')

    def test_machine_ids_list(self):
        """Test getting all machine IDs"""
        self.assertEqual(self.status.machine_ids_list(), ['2', '3'])

    @mock.patch('mojo.juju.status.Juju2Status.status')
    def test_machine_and_container_ids_list(self, status_mock):
        """Test getting all machine and container IDs"""
        # First test on a status that doesn't have containers
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-idle.yaml')
        self.assertEqual(self.status.machine_ids_list(), ['2', '3'])
        self.assertEqual(self.status.machine_and_container_ids_list(), ['2', '3'])
        # And now test on a status that does have containers
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-openstack-with-containers.yaml')
        self.assertEqual(self.status.machine_ids_list(), ['0', '1', '2', '3', '4', '5', '6', '7', '8'])
        self.assertEqual(self.status.machine_and_container_ids_list(),
                         ['0', '0/lxd/0', '0/lxd/1', '0/lxd/2', '0/lxd/3', '0/lxd/4', '0/lxd/5', '0/lxd/6', '0/lxd/7',
                          '1', '1/lxd/0', '1/lxd/1', '1/lxd/2', '1/lxd/3', '1/lxd/4', '1/lxd/5', '1/lxd/6', '1/lxd/7',
                          '2', '3', '4', '5', '6', '7', '8'])

    def test_applications_ips(self):
        """"Test getting services_list"""
        self.assertEqual(self.status.application_ips('apache2'), ['fd28:938b:9119:802d:216:3eff:fe96:aa2d'])
        self.assertEqual(self.status.application_ips('prometheus'), ['10.47.34.203'])

    def test_applications_list(self):
        """"Test getting services_list"""
        self.assertEqual(self.status.applications_list(), ['apache2', 'prometheus'])
        self.assertEqual(self.status.applications_list(exclude_subordinates=False), ['apache2', 'prometheus', 'telegraf'])

    def test_application_machine_numbers(self):
        self.assertEqual(self.status.application_machine_numbers('apache2'), ['3'])

    def test_application_units(self):
        self.assertEqual(self.status.application_units('apache2'), ['apache2/1'])

    @mock.patch('logging.info')
    @mock.patch('mojo.juju.status.wait')
    @mock.patch('mojo.juju.status.Juju2Status.status')
    @mock.patch('subprocess.check_output')
    @mock.patch('time.sleep')
    def test_check_and_wait_juju_wait(self, _sleep, _check_output, status_mock, _wait, _logging_info):
        _wait.return_value = ''
        # Confirm we don't raise a JujuStatusError
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-idle.yaml')
        # First of all call with default (wait_for_steady=False) and confirm
        # wait not called (because juju status is okay).
        self.status.check_and_wait()
        _wait.assert_not_called()
        # Now call with wait_for_steady=True and confirm wait is called even
        # though juju status is okay.
        self.status.check_and_wait(wait_for_steady=True)
        _wait.assert_called_with(max_wait=1800)
        # Also confirm our logging is as expected.
        expected_logging_args_list = [
            "Waiting for environment to reach steady state",
            "Environment has reached steady state",
            ]
        self.assertEquals(len(_logging_info.call_args_list), len(expected_logging_args_list))
        for i, call in enumerate(_logging_info.call_args_list):
            self.assertEquals(call, mock.call(expected_logging_args_list[i]))

    @mock.patch('mojo.juju.status.Juju2Status.status')
    @mock.patch('subprocess.check_output')
    @mock.patch('time.sleep')
    def test_check_and_wait(self, _sleep, _check_output, status_mock):
        # Wait for steady state timeouts
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-busy.yaml')
        with self.assertRaises(mojo.juju.JujuWaitException):
            self.status.check_and_wait(wait_for_steady=True, timeout=1)
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-busy2.yaml')
        with self.assertRaises(mojo.juju.JujuWaitException):
            self.status.check_and_wait(wait_for_steady=True, timeout=1)
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-subordinate-busy.yaml')
        with self.assertRaises(mojo.juju.JujuWaitException):
            self.status.check_and_wait(wait_for_steady=True, timeout=1)
        status_mock.return_value = self._testdata_from_file('juju2-subordinate-busy2.yaml')
        with self.assertRaises(mojo.juju.JujuWaitException):
            self.status.check_and_wait(wait_for_steady=True, timeout=1)

        # error states
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-fail.yaml')
        with self.assertRaises(mojo.juju.JujuStatusError):
            self.status.check_and_wait()
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-subordinate-fail.yaml')
        with self.assertRaises(mojo.juju.JujuStatusError):
            self.status.check_and_wait()

        # And now confirm we don't raise an error on an environment in error
        # state if we pass the right additional_error_states
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status(additional_ready_states=['maintenance', 'executing'])
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        self.status._raw_status = self._testdata_from_file('juju2-busy.yaml')
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        status_mock.return_value = self._testdata_from_file('juju2-busy.yaml')
        with mock.patch('mojo.juju.status.wait', new=lambda max_wait: None):
            self.status.check_and_wait(wait_for_steady=True, timeout=1)

    @mock.patch('subprocess.check_output')
    def test_controller_version(self, mock_check_output):
        with mock.patch('mojo.juju.status.major_version', new=2):
            self.status = mojo.juju.status.Juju2Status()
        mock_check_output.return_value = self._testdata_from_file('juju2-controllers.yaml')
        self.assertEqual('2.0-rc1', self.status.controller_version())

    def test_machine_instance_id(self):
        self.assertEqual(self.status.machine_instance_id('2'), 'juju-fdef5d-2')

    def test_model_version(self):
        self.assertEqual(self.status.model_version(), '2.0-beta18')

    def test_status(self):
        """Test the status function"""
        self.assertEqual(self.status.status(), self.status._raw_status)

        with mock.patch('subprocess.check_output') as mock_check_output:
            # First check with no environment set
            no_env_status = mojo.juju.status.Juju2Status(command_timeout=None)
            no_env_status.status()
            mock_check_output.assert_called_with(
                ['juju', 'status', '--format=yaml'])
            # Second test with an environment set
            env_status = mojo.juju.status.Juju2Status(environment="test-env")
            env_status.status()
            mock_check_output.assert_called_with(
                ['/usr/bin/timeout', '--kill-after', '5', '600',
                 'juju', 'status', '--format=yaml', '-m', 'test-env'])

        with mock.patch('subprocess.check_output') as mock_check_output:
            # Check that if we're calling status() where _raw_status is
            # set, it doesn't query juju status
            self.assertTrue(self.status._raw_status)
            self.status.status()
            self.assertFalse(mock_check_output.called)
            # And now check if we pass the force_update option, that it is
            # called
            self.status.status(force_update=True)
            mock_check_output.assert_called_with(
                ['/usr/bin/timeout', '--kill-after', '5', '600',
                 'juju', 'status', '--format=yaml'])

    @mock.patch('mojo.juju.status.check_output_with_timeout')
    def test_yaml_status(self, _check_output_with_timeout):
        """Test the yaml_status function"""
        _check_output_with_timeout.return_value = self._testdata_from_file('juju2-idle.yaml')
        self.maxDiff = None
        expected = {
            'model': {'name': 'mojo-test', 'controller': 'laptop', 'cloud': 'lxd',
                      'region': 'localhost', 'version': '2.0-beta18'},
            'machines': {'2': {'juju-status': {'current': 'started', 'since': '16 Sep 2016 14:56:46Z',
                                               'version': '2.0-beta18'},
                               'dns-name': '10.47.34.203',
                               'instance-id': 'juju-fdef5d-2',
                               'machine-status': {'current': 'running', 'message': 'Running',
                                                  'since': '16 Sep 2016 14:55:58Z'},
                               'series': 'xenial',
                               'hardware': 'arch=amd64 cpu-cores=0 mem=0M'},
                         '3': {'juju-status': {'current': 'started', 'since': '16 Sep 2016 14:57:11Z',
                                               'version': '2.0-beta18'},
                               'dns-name': 'fd28:938b:9119:802d:216:3eff:fe96:aa2d',
                               'instance-id': 'juju-fdef5d-3',
                               'machine-status': {'current': 'running', 'message': 'Running',
                                                  'since': '16 Sep 2016 14:56:19Z'},
                               'series': 'xenial',
                               'hardware': 'arch=amd64 cpu-cores=0 mem=0M'}},
            'applications': {'apache2': {'charm': 'cs:trusty/apache2-20', 'series': 'xenial',
                                         'os': 'ubuntu', 'charm-origin': 'jujucharms',
                                         'charm-name': 'apache2', 'charm-rev': 20, 'exposed': False,
                                         'application-status': {'current': 'unknown',
                                                                'since': '16 Sep 2016 14:57:53Z'},
                                         'relations': {'juju-info': ['telegraf']},
                                         'units': {'apache2/1': {'workload-status': {'current': 'unknown',
                                                                                     'since': '16 Sep 2016 14:57:53Z'},
                                                                 'juju-status': {'current': 'idle',
                                                                                 'since': '16 Sep 2016 15:20:34Z',
                                                                                 'version': '2.0-beta18'},
                                                                 'machine': '3',
                                                                 'public-address': 'fd28:938b:9119:802d:216:3eff:fe96:aa2d',
                                                                 'subordinates': {'telegraf/0': {
                                                                        'workload-status': {
                                                                            'current': 'active',
                                                                            'message': 'Monitoring apache2/1',
                                                                            'since': '16 Sep 2016 15:23:55Z'},
                                                                        'juju-status': {
                                                                                'current': 'idle',
                                                                                'since': '16 Sep 2016 15:23:55Z',
                                                                                'version': '2.0-beta18'},
                                                                        'upgrading-from': 'cs:~telegraf-charmers/telegraf-6',
                                                                        'open-ports': ['9103/tcp'],
                                                                        'public-address': 'fd28:938b:9119:802d:216:3eff:fe96:aa2d'
                                                                        }}}}},
                             'prometheus': {'charm': 'cs:~prometheus-charmers/prometheus-2', 'series': 'xenial',
                                            'os': 'ubuntu', 'charm-origin': 'jujucharms',
                                            'charm-name': 'prometheus', 'charm-rev': 2, 'exposed': False,
                                            'application-status': {'current': 'active', 'message': 'Ready',
                                                                   'since': '16 Sep 2016 14:59:53Z'},
                                            'relations': {'juju-info': ['telegraf']},
                                            'units': {'prometheus/1': {'workload-status': {'current': 'active',
                                                                                           'message': 'Ready',
                                                                                           'since': '16 Sep 2016 14:59:53Z'},
                                                                       'juju-status': {'current': 'idle',
                                                                                       'since': '16 Sep 2016 15:20:39Z',
                                                                                       'version': '2.0-beta18'},
                                                                       'machine': '2',
                                                                       'open-ports': ['9090/tcp'],
                                                                       'public-address': '10.47.34.203',
                                                                       'subordinates': {'telegraf/1': {
                                                                            'workload-status': {
                                                                                'current': 'active',
                                                                                'message': 'Monitoring prometheus/1',
                                                                                'since': '16 Sep 2016 15:23:55Z'},
                                                                            'juju-status': {'current': 'idle',
                                                                                            'since': '16 Sep 2016 15:23:55Z',
                                                                                            'version': '2.0-beta18'},
                                                                            'upgrading-from': 'cs:~telegraf-charmers/telegraf-6',
                                                                            'open-ports': ['9103/tcp'],
                                                                            'public-address': '10.47.34.203'}}}}},
                             'telegraf': {'charm': 'cs:~telegraf-charmers/telegraf-6',
                                          'series': 'xenial', 'os': 'ubuntu', 'charm-origin': 'jujucharms',
                                          'charm-name': 'telegraf', 'charm-rev': 6, 'exposed': False,
                                          'application-status': {'current': 'active',
                                                                 'message': 'Monitoring apache2/1',
                                                                 'since': '16 Sep 2016 15:23:55Z'},
                                          'relations': {'juju-info': ['apache2', 'prometheus']},
                                          'subordinate-to': ['apache2', 'prometheus']}}}
        self.assertEqual(self.status.yaml_status(force_update=True), expected)


class Juju2CheckOutputWithTimeoutTestException(Exception):
    pass


class Juju2CheckOutputWithTimeoutTest(TestCase):
    """
    Test that check_output_with_timeout handles timeout(1)'s behaviour as documented, i.e.:

        If the command times out, and --preserve-status is not set, then exit with
        status 124.  Otherwise, exit with the status of COMMAND.  If no signal
        is specified, send the TERM signal upon timeout.  The TERM signal kills
        any process that does not block or catch that signal.  It may be necessary
        to use the KILL (9) signal, since this signal cannot be caught, in which
        case the exit status is 128+9 rather than 124."""

    @mock.patch('subprocess.check_output')
    def test_command_succeeds(self, check_output_mock):
        """command succeeds within timeout -> returns output, no exception"""

        command = ['/bin/echo', 'blorp']
        check_output_mock.return_value = 'blorp'

        self.assertEqual(
            mojo.juju.status.check_output_with_timeout(
                command,
                command_timeout=5,
                timeout_exception=Juju2CheckOutputWithTimeoutTestException),
            'blorp')

    @mock.patch('subprocess.check_output')
    def test_command_fails_within_timeout(self, check_output_mock):
        """command fails within timeout -> CalledProcessError thrown"""

        import subprocess

        command = ['/bin/echo', 'blorp']
        check_output_mock.side_effect = subprocess.CalledProcessError(returncode=1, cmd=command)

        with self.assertRaises(subprocess.CalledProcessError) as arcm:
            mojo.juju.status.check_output_with_timeout(
                command,
                command_timeout=5,
                timeout_exception=Juju2CheckOutputWithTimeoutTestException)

        self.assertEqual(arcm.exception.returncode, 1)

    @mock.patch('subprocess.check_output')
    def test_command_times_out_and_is_terminated(self, check_output_mock):
        """command times out, SIGTERM exits -> CalledProcessError.returncode==124 -> timeout_exception('...')"""

        import subprocess

        command = ['/bin/echo', 'blorp']
        check_output_mock.side_effect = subprocess.CalledProcessError(returncode=124, cmd=command)

        with self.assertRaises(Juju2CheckOutputWithTimeoutTestException) as arcm:
            mojo.juju.status.check_output_with_timeout(
                command,
                command_timeout=5,
                timeout_exception=Juju2CheckOutputWithTimeoutTestException)

        self.assertTrue(
            arcm.exception.message.endswith(' timed out after 5 seconds'))

    @mock.patch('subprocess.check_output')
    def test_command_times_out_and_is_killed(self, check_output_mock):
        """command times out, SIGKILL exits -> CalledProcessError.returncode==128+9 -> timeout_exception('... (killed)')"""

        import subprocess

        command = ['/bin/echo', 'blorp']
        check_output_mock.side_effect = subprocess.CalledProcessError(returncode=128+9, cmd=command)

        with self.assertRaises(Juju2CheckOutputWithTimeoutTestException) as arcm:
            mojo.juju.status.check_output_with_timeout(
                command,
                command_timeout=5,
                timeout_exception=Juju2CheckOutputWithTimeoutTestException)

        self.assertTrue(
            arcm.exception.message.endswith(' timed out after 5 seconds (killed)'))


class Juju2ChecksTest(TestCase):
    def setUp(self):
        self.testdata = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'testdata')
        self.status_path = os.path.join(self.testdata, 'juju2-idle2.yaml')

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_Errors_success(self, mock_stdout):
        sys.argv = ['self', '-f', self.status_path, 'errors']
        return_code = check()
        self.assertEqual(0, return_code)
        self.assertEqual(mock_stdout.getvalue(), 'OK: No Juju status errors found.\n')

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_Errors_failure(self, mock_stdout):
        error_status = os.path.join(self.testdata, 'juju2-fail.yaml')
        sys.argv = ['self', '-f', error_status, 'errors']
        return_code = check()
        self.assertEqual(2, return_code)
        self.assertEqual(mock_stdout.getvalue(), 'CRITICAL: apache2 is in status error\n')

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_Unused_success(self, mock_stdout):
        sys.argv = ['self', '-f', self.status_path, 'unused_machines']
        return_code = check()
        self.assertEqual(0, return_code)
        self.assertEqual(mock_stdout.getvalue(), 'OK: All machines are being used.\n')

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_Unused_failure(self, mock_stdout):
        unused_status = os.path.join(self.testdata, 'juju2-unused-machines.yaml')
        sys.argv = ['self', '-f', unused_status, 'unused_machines']
        return_code = check()
        self.assertEqual(2, return_code)
        self.assertEqual(mock_stdout.getvalue(), "CRITICAL: Unused machines found: ['3']\n")

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_UpgradeAvailable_success(self, mock_stdout):
        sys.argv = ['self', '-f', self.status_path, 'upgrade_available', '--check_version', '2.0-rc1']
        return_code = check()
        self.assertEqual(0, return_code)
        self.assertEqual(mock_stdout.getvalue(), 'OK: The Juju controller and model are running the same version (2.0-rc1).\n')

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_UpgradeAvailable_failure(self, mock_stdout):
        sys.argv = ['self', '-f', self.status_path, 'upgrade_available', '--check_version', '2.0-rc2']
        return_code = check()
        self.assertEqual(2, return_code)
        self.assertEqual(mock_stdout.getvalue(), 'CRITICAL: controller version 2.0-rc2 != model version 2.0-rc1\n')


class Juju2ParseStatusTest(TestCase):
    def setUp(self):
        self.testdata = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'testdata')
        self.status_path = os.path.join(self.testdata, 'juju2-idle3.yaml')

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_GetIPs_all(self, mock_stdout):
        sys.argv = ['self', '-f', self.status_path, 'get_ips', 'all']
        rc = parse_status()
        self.assertEqual(rc, 0)
        self.assertEqual(mock_stdout.getvalue(), '10.25.2.111\n10.242.242.242\n10.25.2.109\n10.25.2.112\n10.25.2.110\n')

    @mock.patch('sys.stdout', new_callable=StringIO)
    def test_GetIPs_app(self, mock_stdout):
        sys.argv = ['self', '-f', self.status_path, 'get_ips', 'ubuntu']
        rc = parse_status()
        self.assertEqual(rc, 0)
        self.assertEqual(mock_stdout.getvalue(), '10.25.2.109\n10.25.2.111\n10.25.2.110\n')
