# Copyright 2014-2018 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import mock

from unittest import TestCase

from mojo.manifest import Manifest
from mojo.exceptions import InvalidManifestException, ConfigNotFoundException
from mojo.project import Project
from mojo.tests.utils import MojoTestCaseMixin


test_manifests_for_includes = [{
    'name': "An include at the end of a file",
    'expected_phases': ['collect', 'repo', 'deploy', 'script', 'verify'],
    'manifests': {
        'manifest': """
collect
repo
deploy config=services
include config=other-manifest
""",
        'other-manifest': """
script config=post-deploy
verify
""",
        },
    }, {
    'name': "An include in the middle of a file",
    'expected_phases': ['collect', 'repo', 'deploy', 'script', 'verify'],
    'manifests': {
        'manifest': """
collect
repo
deploy config=services
include config=other-manifest
verify
""",
        'other-manifest': """
script config=post-deploy
""",
        }
    }, {
    'name': "Nested include files",
    'expected_phases': [
        'collect',  # manifest
        'repo',
        'deploy',   # deploy-manifest
        'verify',   # verify-manifest
        'script',
        'verify',
        'deploy',   # deploy-manifest
        'script',   # manifest
        ],
    'manifests': {
        'manifest': """
collect
repo
include config=deploy-manifest
script
""",
        'deploy-manifest': """
deploy config=services
include config=verify-manifest
deploy config=somethingelse
""",
        'verify-manifest': """
verify
script config=other-test
verify config=verify-nagios
""",
        },
    }, {
    'name': "Nested include files in subdir",
    'expected_phases': [
        'collect',  # manifest
        'repo',
        'deploy',   # deploy-manifest
        'verify',   # verify-manifest
        'script',
        'verify',
        'deploy',   # deploy-manifest
        'script',   # manifest
        ],
    'manifests': {
        'manifest': """
collect
repo
include config=manifests/deploy-manifest
script
""",
        'manifests/deploy-manifest': """
deploy config=services
include config=manifests/verify-manifest
deploy config=somethingelse
""",
        'manifests/verify-manifest': """
verify
script config=other-test
verify config=verify-nagios
""",
        },
    }]

test_manifests_for_missing_include = [{
    'name': "Include file doesn't exist",
    'manifests': {
        'manifest': """
collect
repo
deploy config=services
include config=non-existent-manifest
""",
        'other-manifest': """
script config=post-deploy
verify
""",
        },
    }]

test_manifests_for_invalid_phase = [{
    'name': "Invalid phase name",
    'manifests': {
        'manifest': """
explode
repo
deploy config=services
include config=other-manifest
""",
        'other-manifest': """
script config=post-deploy
verify
""",
        },
    }]

test_manifests_for_invalid_arg = [{
    'name': "Invalid phase argument",
    'manifests': {
        'manifest': """
repo
deploy notavalidarg
"""},
    }]


test_manifests_for_multiple_args = [{
    'name': "Multiple phase arguments",
    'manifests': {
        'manifest': """
repo
deploy config=foo config=bar
"""},
    }]


test_manifests_for_comment_and_include_ordering = [{
    'name': "Comment ordering w/r/t includes",
    'expected_phases': ['collect', 'comment', 'deploy', 'comment', 'verify'],
    'manifests': {
        'manifest': """
collect
# Deploy the environment.
include config=manifest-deploy
# Verify the environment.
include config=manifest-verify
""",
        'manifest-deploy': """
deploy
""",
        'manifest-verify': """
verify
""",
        },
    }]


class ManifestTestCase(TestCase, MojoTestCaseMixin):

    def test_include_sets_correct_phases(self):
        for example in test_manifests_for_includes:
            spec = self.create_spec(manifests=example['manifests'])

            manifest = Manifest(spec=spec, manifest_file='manifest',
                                stage='devel')

            actual_phases = [phase.name for phase in manifest.phases]
            self.assertEqual(
                example['expected_phases'], actual_phases,
                "The example '{}' expected the phases:\n{}\n "
                "but had\n{}\ninstead.".format(
                    example['name'], example['expected_phases'],
                    actual_phases))

    def test_failing_manifests_missing_include(self):
        for example in test_manifests_for_missing_include:
            spec = self.create_spec(manifests=example['manifests'])

            with self.assertRaises(ConfigNotFoundException):
                Manifest(spec=spec, manifest_file='manifest',
                         stage='devel')

    def test_failing_manifests_invalid_phase(self):
        for example in test_manifests_for_invalid_phase:
            spec = self.create_spec(manifests=example['manifests'])

            with self.assertRaises(InvalidManifestException):
                Manifest(spec=spec, manifest_file='manifest',
                         stage='devel')

    def test_failing_manifests_invalid_arg(self):
        for example in test_manifests_for_invalid_arg:
            spec = self.create_spec(manifests=example['manifests'])

            with self.assertRaises(InvalidManifestException):
                Manifest(spec=spec, manifest_file='manifest',
                         stage='devel')

    def test_failing_manifests_multiple_args(self):
        for example in test_manifests_for_multiple_args:
            spec = self.create_spec(manifests=example['manifests'])

            with self.assertRaises(InvalidManifestException):
                Manifest(spec=spec, manifest_file='manifest',
                         stage='devel')

    def test_manifest_comment_parsing(self):
        manifest_text = """
collect
## Should be ignored.
repo
# Here we go
deploy config=services
# Done
# with a multiline comment.
script config=post-deploy
# Verify
verify
"""
        spec = self.create_spec(manifests=dict(manifest=manifest_text))
        manifest = Manifest(spec=spec, manifest_file='manifest',
                            stage='devel')

        actual_phases = [phase.name for phase in manifest.phases]
        self.assertEqual([
            'collect',
            'repo',
            'comment',
            'deploy',
            'comment',
            'script',
            'comment',
            'verify',
            ], actual_phases)
        self.assertEqual("Here we go", manifest.phases[2].options['text'])
        self.assertEqual("Done\nwith a multiline comment.",
                         manifest.phases[4].options['text'])
        self.assertEqual("Verify", manifest.phases[6].options['text'])

    def test_manifest_comment_and_include_ordering(self):

        for example in test_manifests_for_comment_and_include_ordering:
            spec = self.create_spec(manifests=example['manifests'])

            manifest = Manifest(spec=spec, manifest_file='manifest',
                                stage='devel')

            actual_phases = [phase.name for phase in manifest.phases]
            self.assertEqual(
                example['expected_phases'], actual_phases,
                "The example '{}' expected the phases:\n{}\n "
                "but had\n{}\ninstead.".format(
                    example['name'], example['expected_phases'],
                    actual_phases))

    @mock.patch('__builtin__.raw_input')
    @mock.patch('logging.info')
    def test_manifest_interactive(self, _logging_info, _mock_raw_input):
        manifest_list = ["collect", "deploy config=services"]
        spec = self.create_spec(manifests=dict(manifest="\n".join(manifest_list)))
        manifest = Manifest(spec=spec, manifest_file='manifest', stage='devel')
        project = Project('unused', 'trusty', '/srv/mojo')
        workspace = self.create_workspace()
        # Test with Stop as the directive first. We expect it to prompt for
        # the first phase, and then stop.
        _mock_raw_input.return_value = 's'
        expected_call_args_list = [
            "Interactive mode... about to run:",
            "  collect ",
            "",
            "[(Y)es] / (N)o / (S)top",
            ]
        manifest.run(project, workspace, stage='devel', interactive=True)
        self.assertEquals(len(_logging_info.call_args_list), len(expected_call_args_list))
        for i, call in enumerate(_logging_info.call_args_list):
            self.assertEquals(call, mock.call(expected_call_args_list[i]))
        _logging_info.reset_mock()
        # And now test with No as the directive. We expect it to prompt for
        # each phase, but not execute.
        _mock_raw_input.return_value = 'n'
        expected_call_args_list = [
            "Interactive mode... about to run:",
            "  collect ",
            "",
            "[(Y)es] / (N)o / (S)top",
            "Skipping...",
            "",
            "Interactive mode... about to run:",
            "  deploy config=services",
            "",
            "[(Y)es] / (N)o / (S)top",
            "Skipping...",
            "",
            ]
        manifest.run(project, workspace, stage='devel', interactive=True)
        self.assertEquals(len(_logging_info.call_args_list), len(expected_call_args_list))
        for i, call in enumerate(_logging_info.call_args_list):
            self.assertEquals(call, mock.call(expected_call_args_list[i]))
