# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import commands
import json
import mock
import os
import subprocess
import sys
import time

from shutil import rmtree
from tempfile import mkdtemp
from unittest import TestCase

import mojo.project
import mojo.spec

from mojo.utils import chdir, is_config_line
from mojo.tests.utils import MojoTestCaseMixin

# Make sure that the local mojo is importable
import mojo.workspace

local_copy = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "..",
    "mojo")
if os.path.exists(local_copy) and os.path.isdir(local_copy):
    sys.path.insert(0, local_copy)

try:
    import mojo
except ImportError:
    raise ImportError("Could not import local copy of mojo.")


class WorkspaceTest(TestCase):

    def setUp(self):
        super(WorkspaceTest, self).setUp()
        self.bad_path = "/INVALID/PATH"
        self.good_path = mkdtemp()
        self.addCleanup(rmtree, self.good_path)
        self.working_spec_path = mkdtemp()
        self.addCleanup(rmtree, self.working_spec_path)

        self.build_dir_name = "build"
        self.local_dir_name = "local"
        self.log_dir_name = "log"
        self.spec_dir_name = "spec"
        self.all_names = [
            self.build_dir_name,
            self.local_dir_name,
            self.log_dir_name,
            self.spec_dir_name,
            ]

    def test_workspace_sets_path(self):
        w = mojo.workspace.Workspace("workspace_name", self.good_path)
        self.assertEqual(w.path, self.good_path)

        with self.assertRaises(mojo.workspace.NoSuchWorkspaceError):
            w = mojo.workspace.Workspace("workspace_name", self.bad_path)

    def test_workspace_sets_default_path(self):
        with chdir(self.good_path):
            w = mojo.workspace.Workspace("workspace_name", None)
        self.assertEqual(w.path, self.good_path)

    def test_prepares_path(self):
        w = mojo.workspace.Workspace("workspace_name", self.good_path)
        self.assertTrue(w.prepare())

        all_subdirs = [os.path.join(self.good_path, d) for d in self.all_names]
        for subdir in all_subdirs:
            if os.path.basename(subdir) == 'spec':  # Skip the spec dir
                continue
            self.assertTrue(os.path.exists(subdir))

    def test_prepares_new_path(self):
        w = mojo.workspace.Workspace("workspace_name", self.good_path)
        self.assertTrue(w.prepare())

        all_subdirs = [os.path.join(self.good_path, d) for d in self.all_names]
        for subdir in all_subdirs:
            if os.path.basename(subdir) == 'spec':  # Skip the spec dir
                continue
            self.assertTrue(os.path.exists(subdir))

    def test_prepare_errors_return_false(self):
        os.chmod(self.good_path, 0o555)
        w = mojo.workspace.Workspace("workspace_name", self.good_path)
        self.assertFalse(w.prepare())

    def test_returns_dirs(self):
        w = mojo.workspace.Workspace("workspace_name", self.good_path)

        build_dir = os.path.join(self.good_path, self.build_dir_name)
        self.assertEqual(w.build_dir, build_dir)

        local_dir = os.path.join(self.good_path, self.local_dir_name)
        self.assertEqual(w.local_dir, local_dir)

        log_dir = os.path.join(self.good_path, self.log_dir_name)
        self.assertEqual(w.log_dir, log_dir)

        spec_dir = os.path.join(self.good_path, self.spec_dir_name)
        self.assertEqual(w.spec_dir, spec_dir)

    @mock.patch('logging.info')
    def test_workspace_spec_update_from_source(self, _logging_info):
        w = mojo.workspace.Workspace("workspace_name", self.good_path, spec_url=self.working_spec_path)
        w.spec.update_from_source()
        # Confirm the working spec dir is empty and source for spec only
        # contains the auto-generated codetree-collect-info yaml file.
        self.assertEqual(os.listdir(self.working_spec_path), [])
        self.assertEqual(os.listdir(w.spec_dir), ['codetree-collect-info.yaml'])
        # Now create a file in our source spec dir.
        with open(os.path.join(self.working_spec_path, 'testfile'), 'w') as f:
            f.write("testcontents")
        # Refresh the working spec directory from source.
        w.spec.update_from_source()
        # And now confirm our working spec directory contains the file.
        self.assertEqual(sorted(os.listdir(w.spec_dir)), ['codetree-collect-info.yaml', 'testfile'])
        with open(os.path.join(w.spec_dir, 'testfile'), 'r') as f:
            self.assertEqual(f.read(), 'testcontents')


class SpecTest(TestCase, MojoTestCaseMixin):

    def setUp(self):
        self.git_daemon_pid = None

    def test_get_config(self):
        cases = [{
            "config_files": ['staging/service.yml'],
            "input": dict(base="service.yml", stage='staging'),
            "expected": "staging/service.yml",
            }, {
            "config_files": ['service.yml', 'staging/service.yml'],
            "input": dict(base="service.yml", stage='staging'),
            "expected": "staging/service.yml",
            }, {
            "config_files": ['service.yml', 'service.yml-staging'],
            "input": dict(base="service.yml", stage=None),
            "expected": "service.yml",
            }, {
            "config_files": [
                'service.yml',
                'pes/service.yml',
                'pes/staging/service.yml',
                ],
            "input": dict(base="service.yml", stage='pes/staging'),
            "expected": "pes/staging/service.yml",
            }, {
            "config_files": ['service.yml', 'pes/service.yml'],
            "input": dict(base="service.yml", stage='pes/staging'),
            "expected": "pes/service.yml",
            }]
        for test_case in cases:
            spec_dir = self.create_spec_dir(test_case['config_files'],
                                            stage=test_case['input']['stage'])

            spec = mojo.spec.Spec(spec_dir, spec_dir)  # Tests spec_dir and spec_url being the same

            expected_config = os.path.join(spec_dir, test_case["expected"])
            actual_config = spec.get_config(**test_case['input'])
            self.assertEqual(expected_config, actual_config)

    def test_get_configs(self):
        cases = [{
            "config_files": ['staging/service.yml'],
            "input": dict(base="service.yml", stage='staging'),
            "expected": ["staging/service.yml"],
            }, {
            "config_files": ['service.yml', 'staging/service.yml'],
            "input": dict(base="service.yml", stage='staging'),
            "expected": ["service.yml", "staging/service.yml"],
            }, {
            "config_files": [
                'staging/service.yml',
                'service.yml',
                ],
            "input": dict(base="service.yml", stage='staging'),
            "expected": [
                "service.yml",
                'staging/service.yml',
                ],
            }, {
            "config_files": ['service.yml', 'service.yml-staging'],
            "input": dict(base="service.yml", stage=None),
            "expected": ["service.yml"],
            }, {
            "config_files": [
                'service.yml',
                'pes/service.yml',
                'pes/staging/service.yml',
                ],
            "input": dict(base="service.yml", stage='pes/staging'),
            "expected": [
                "service.yml",
                "pes/service.yml",
                "pes/staging/service.yml",
                ],
            }, {
            "config_files": ['service.yml', 'pes/service.yml'],
            "input": dict(base="service.yml", stage='pes/staging'),
            "expected": ["service.yml", "pes/service.yml"],
            }]
        for test_case in cases:
            spec_dir = self.create_spec_dir(test_case['config_files'],
                                            stage=test_case['input']['stage'])

            spec = mojo.spec.Spec(spec_dir)  # Tests spec_url being None

            actual_configs = spec.get_configs(**test_case['input'])
            self.assertEqual(test_case["expected"], actual_configs)

    def _wait_for_socket(self, addr, port):
        socket = '{}:{}'.format(addr, port)
        cmd = ['ss', 'state', 'listening', 'src', socket]
        waited = 1

        while waited < 10:
            time.sleep(waited)
            waited = waited * 2
            output = subprocess.check_output(cmd)
            if output.count(socket) > 0:
                return True

        return False

    def test_git_spec(self):
        # Do the spec_dir setup like other tests as our baseline
        bzr_spec_dir = self.create_spec_dir(['service.yml', 'staging/service.yml'], stage='staging')
        bzr_spec = mojo.spec.Spec(bzr_spec_dir)
        bzr_configs = bzr_spec.get_configs(base="service.yml", stage="staging")

        # Copy the bzr_spec_dir, turn into a git repo and serve at a local git daemon
        # A bunch of this was stolen from the codetree git testing
        git_spec_dir = mkdtemp()
        self.addCleanup(rmtree, git_spec_dir)
        os.rmdir(git_spec_dir)
        git_dir = mkdtemp()
        self.addCleanup(rmtree, git_dir)
        os.rmdir(git_dir)
        subprocess.check_call(['cp', '-r', bzr_spec_dir, git_dir])

        # give git an identity if none exists
        if subprocess.call(['git', 'config', '--global', '--get', 'user.email']) == 1:
            subprocess.call(['git', 'config', '--global', 'user.email', "you@example.com"])
        if subprocess.call(['git', 'config', '--global', '--get', 'user.name']) == 1:
            subprocess.call(['git', 'config', '--global', 'user.name', "Your Name"])

        subprocess.check_call(['git', '-C', git_dir, 'init', '-q'])
        subprocess.check_call(('git', '-C', git_dir, 'add', '*'))
        subprocess.check_call(('git', '-C', git_dir, 'commit', '-q', '-a', '-m', 'initial commit'))

        self.git_daemon_pid = subprocess.Popen(('git', '-C', git_dir, 'daemon', '--reuseaddr', '--export-all',
                                                '--listen=127.0.0.1', '--base-path={}'.format(git_dir)))
        git_spec_url = "git://localhost/.git"

        self._wait_for_socket('127.0.0.1', 9418)
        git_spec = mojo.spec.Spec(git_spec_dir, git_spec_url)
        git_configs = git_spec.get_configs(base="service.yml", stage="staging")

        self.assertEqual(bzr_configs, git_configs)

    def tearDown(self):
        if self.git_daemon_pid is not None:
            self.git_daemon_pid.terminate()


class UtilsTest(TestCase):

    def test_bicommand(self):
        """bicommand() produces the same output as getstatusoutput()."""
        cmds = ["ls -1 /tmp",  # rarely will fail, will usually be fast
                "ls ~",  # never (?) will fall, will usually be fast
                "uname",  # shouldn't ever fail, will be fast
                "find /this/file/should/not/exist",  # should be stderr
                "find /usr/share/doc/dpkg",  # will be slow
                ]
        for cmd in cmds:
            bc_status, bc_output = mojo.utils.bicommand(cmd)
            c_status, c_output = commands.getstatusoutput(cmd)

            # We don't care about return codes, we just want to confirm they're
            # both returning zero or non-zero
            if bc_status != 0:
                bc_status = 1
            if c_status != 0:
                c_status = 1

            self.assertEqual(bc_status, c_status)
            self.assertEqual(bc_output, c_output)

    def test_bicommand_environment_expansion(self):
        """Environment variables in bicommand invocations are expanded."""
        response = "I've come to talk with you again."
        status, output = mojo.utils.bicommand(
            "echo ${HELLO_DARKNESS_MY_OLD_FRIEND}",
            env={"HELLO_DARKNESS_MY_OLD_FRIEND": response},
            )
        self.assertEqual(status, 0)
        self.assertEqual(output, response)

    def test_is_config_line(self):
        """Test various types of lines for whether they are valid config or not"""
        self.assertEqual(True, is_config_line('This is some config'))
        self.assertEqual(True, is_config_line(' This is some config '))
        self.assertEqual(False, is_config_line(' '))
        self.assertEqual(False, is_config_line('# This is not config'))
        self.assertEqual(False, is_config_line(' # This is not config'))


class ProjectTest(TestCase):

    def test_project_lxc_trusty(self):
        project = mojo.project.Project('my-project', 'trusty', '/srv/mojo', container_class=mojo.contain.LXCContainer)
        self.assertTrue(type(project.container) == mojo.contain.LXCContainer)
        self.assertTrue(project.container.container_root == '/srv/mojo/my-project/trusty/ROOTFS')

    def test_project_lxc_trusty_exists(self):
        """Pre-existing project has lxc root at
        /var/lib/lxc/my-project.trusty/rootfs"""
        mojo_root = mkdtemp()
        self.addCleanup(rmtree, mojo_root)
        project_series_dir = os.path.join(mojo_root, 'my-project', 'trusty')
        mojo.utils.mkdirs(project_series_dir)
        project_info_file = os.path.join(mojo_root, 'my-project', 'trusty', '.project')
        existing_project_info = {'container_root': '/var/lib/lxc/my-project.trusty/rootfs',
                                 'container_class': 'lxc'}
        with open(project_info_file, 'w') as info:
            info.write(json.dumps(existing_project_info))

        project = mojo.project.Project('my-project', 'trusty', mojo_root, container_class=mojo.contain.LXCContainer)
        self.assertTrue(type(project.container) == mojo.contain.LXCContainer)
        self.assertTrue(project.container.container_root == '/var/lib/lxc/my-project.trusty/rootfs')

    def test_project_containerless(self):
        project = mojo.project.Project('my-project', 'trusty', '/srv/mojo', container_class=mojo.contain.NoContainer)
        self.assertTrue(type(project.container) == mojo.contain.NoContainer)
        self.assertTrue(project.container.container_root == '/srv/mojo/my-project/trusty')

    def test_project_containerless_exists(self):
        """Pre-existing containerless project"""
        mojo_root = mkdtemp()
        self.addCleanup(rmtree, mojo_root)
        project_series_dir = os.path.join(mojo_root, 'my-project', 'trusty')
        mojo.utils.mkdirs(project_series_dir)
        project_info_file = os.path.join(project_series_dir, '.project')
        existing_project_info = {'container_root': project_series_dir,
                                 'container_class': 'containerless'}
        with open(project_info_file, 'w') as info:
            info.write(json.dumps(existing_project_info))

        # Recieves the default LXCContainer but is a pre-existing
        # NoContainer project
        project = mojo.project.Project('my-project', 'trusty', mojo_root, container_class=mojo.contain.LXCContainer)
        self.assertTrue(type(project.container) == mojo.contain.NoContainer)
        self.assertTrue(project.container.container_root == project_series_dir)
