# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import mock
import os
import shutil
import stat
import tempfile
from textwrap import dedent
from unittest import TestCase

from deployer.utils import ErrorExit
from mojo.phase import (
    process_config_changes_from_diff,
    BundlePhase,
    CharmAuditPhase,
    CharmRepoPhase,
    CollectPhase,
    CollectPhaseException,
    DeployerPhase,
    FileDoesNotExistException,
    NagiosCheckPhase,
    NagiosCheckPhaseException,
    SecretsPhase,
    )
from mojo.tests import utils
from mojo.project import Project


class DeployerPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    @mock.patch('mojo.phase.bicommand')
    @mock.patch('mojo.phase.DeployerPhase._confirm_charm_repo')
    @mock.patch('mojo.phase.DeployerPhase._only_charmstore_charms')
    @mock.patch('mojo.phase.DeployerPhase._get_target')
    @mock.patch('mojo.juju.Status.check_and_wait')
    @mock.patch('mojo.phase.DeployerPhase._get_command')
    def get_deployer_command_for_run(self, phase, workspace, stage, _get_command, _check_wait,
                                     _get_target, _only_charmstore_charms, _confirm_charm_repo, _bicommand):
        _only_charmstore_charms.return_value = False
        _get_target.return_value = "testTarget"
        _bicommand.return_value = (0, 'fake-output')
        _get_command.return_value = "juju-deployer"

        project = Project('unused', 'trusty', '/srv/mojo')
        phase.run(project, workspace, stage)

        return _bicommand.call_args[0][0]

    @mock.patch('subprocess.call')
    @mock.patch('mojo.phase.DeployerPhase._confirm_charm_repo')
    @mock.patch('mojo.phase.DeployerPhase._only_charmstore_charms')
    @mock.patch('mojo.phase.DeployerPhase._get_target')
    @mock.patch('mojo.juju.Status.check_and_wait')
    @mock.patch('mojo.phase.DeployerPhase._get_command')
    def get_deployer_command_for_diff(self, phase, workspace, stage, _get_command, _check_wait,
                                      _get_target, _only_charmstore_charms, _confirm_charm_repo, _call):
        _only_charmstore_charms.return_value = False
        _get_target.return_value = "testTarget"
        _call.return_value = (0)
        _get_command.return_value = "juju-deployer"

        project = Project('unused', 'trusty', '/srv/mojo')
        phase.diff(project, workspace, stage)

        mock_command = ""
        for arg in _call.call_args[0][0]:
            mock_command += "{} ".format(arg)

        return mock_command

    @mock.patch('subprocess.call')
    @mock.patch('mojo.phase.DeployerPhase._confirm_charm_repo')
    @mock.patch('mojo.phase.DeployerPhase._only_charmstore_charms')
    @mock.patch('mojo.phase.DeployerPhase._get_target')
    @mock.patch('mojo.juju.Status.check_and_wait')
    @mock.patch('mojo.phase.DeployerPhase._get_command')
    def get_deployer_output_for_show(self, phase, workspace, stage, target, _get_command, _check_wait,
                                     _get_target, _only_charmstore_charms, _confirm_charm_repo, _call):
        _only_charmstore_charms.return_value = False
        _get_target.return_value = target
        _call.return_value = (0)
        _get_command.return_value = "juju-deployer"

        project = Project('unused', 'trusty', '/srv/mojo')
        return phase.show(project, workspace, stage, return_output=True)

    def test_default_config_rendered_for_deployer(self):
        phase = DeployerPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/deploy', 'deploy'],
            stage='staging')

        command = self.get_deployer_command_for_run(phase, workspace, 'staging')

        # Verify that the configs were both rendered.
        self.assertTrue(os.path.exists(
            '{}/deploy'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/deploy'.format(workspace.repo_dir)))
        # Verify that the configs were included in the deployer command.
        expected_command = (
            'juju-deployer -c {configs_dir}/deploy '
            '-c {configs_dir}/staging/deploy '
            '-s 0 -t 2700 -d testTarget -W -u'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

    def test_multiple_configs_rendered_for_deployer(self):
        phase = DeployerPhase(config='services.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'services.cfg'],
            stage='staging')

        command = self.get_deployer_command_for_run(phase, workspace, 'staging')

        # Verify that the configs were both rendered.
        self.assertTrue(os.path.exists(
            '{}/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/services.cfg'.format(workspace.repo_dir)))
        # Verify that the configs were included in the deployer command.
        expected_command = (
            'juju-deployer -c {configs_dir}/services.cfg '
            '-c {configs_dir}/staging/services.cfg '
            '-s 0 -t 2700 -d testTarget -W -u'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

    def test_diff_for_deployer(self):
        phase = DeployerPhase(config='services.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'services.cfg'],
            stage='staging')

        command = self.get_deployer_command_for_diff(phase, workspace, 'staging')

        expected_command = (
            'juju-deployer -c {configs_dir}/services.cfg '
            '-c {configs_dir}/staging/services.cfg --diff testTarget '.format(configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

    @mock.patch('subprocess.check_call')
    @mock.patch('logging.info')
    def test_process_config_changes_from_diff(self, _logging_info, _check_call):
        # First confirm with no diff, check_call isn't invoked, even with
        # preview=False.
        deployer_diff = ''
        process_config_changes_from_diff(deployer_diff, preview=False)
        self.assertFalse(_logging_info.called)
        self.assertFalse(_check_call.called)
        # Now check with a diff, but preview=True (default), we log but don't
        # do anything.
        deployer_diff = open('mojo/tests/testdata/deployer-diff.yaml', 'r').read()
        process_config_changes_from_diff(deployer_diff)
        expected_logging = """Preview config change (NOT RUN):

juju config nrpe nagios_master=None hostgroups= export_nagios_definitions=False procs= swap=-w 40% -c 25%"""
        self.assertEqual(_logging_info.call_args, mock.call(expected_logging))
        self.assertFalse(_check_call.called)
        # Now check with a diff and preview=False we log but also execute our
        # command.
        _logging_info.reset_mock()
        deployer_diff = open('mojo/tests/testdata/deployer-diff.yaml', 'r').read()
        process_config_changes_from_diff(deployer_diff, preview=False)
        expected_logging = """Running config change for items:

nagios_master hostgroups export_nagios_definitions procs swap"""
        self.assertEqual(_logging_info.call_args, mock.call(expected_logging))
        expected_subprocess = ["juju", "config", "nrpe", "nagios_master=None", "hostgroups=",
                               "export_nagios_definitions=False", "procs=", "swap=-w 40% -c 25%"]
        self.assertEqual(_check_call.call_args, mock.call(expected_subprocess))
        # Now test with a different diff output.
        _logging_info.reset_mock()
        _check_call.reset_mock()
        deployer_diff = open('mojo/tests/testdata/deployer-diff2.yaml', 'r').read()
        process_config_changes_from_diff(deployer_diff, preview=False)
        expected_logging = """Running config change for items:

users zombies"""
        self.assertEqual(_logging_info.call_args, mock.call(expected_logging))
        expected_subprocess = ['juju', 'config', 'nrpe', 'users=-w 20 -c 25', 'zombies=-w 3 -c 6 -s Z']
        self.assertEqual(_check_call.call_args, mock.call(expected_subprocess))

    def test_show_for_deployer(self):
        phase = DeployerPhase(config='services.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'services.cfg'],
            stage='staging')

        # This is generated in mojo/tests/utils.py as part of the
        # create_unique_bzr_branch function, which is called by the
        # create_spec_dir function.
        expected = "series: trusty\nservices:\n  apache2: {{charm: {}/trusty/apache2}}\n".format(workspace.repo_dir)
        actual = self.get_deployer_output_for_show(phase, workspace, 'staging', 'deploy-config')
        self.assertEqual(actual, expected)

        # Now test with a target that doesn't exist.
        with self.assertRaises(ErrorExit):
            self.get_deployer_output_for_show(phase, workspace, 'staging', 'deploy')

    def test_comma_separated_configs_rendered_for_deployer(self):
        phase = DeployerPhase(config='services.cfg,placement.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'staging/placement.cfg', 'services.cfg'],
            stage='staging')

        command = self.get_deployer_command_for_run(phase, workspace, 'staging')

        # Verify that the configs were rendered.
        self.assertTrue(os.path.exists(
            '{}/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/placement.cfg'.format(workspace.repo_dir)))
        # Verify that the configs were included in the deployer command.
        expected_command = (
            'juju-deployer -c {configs_dir}/services.cfg '
            '-c {configs_dir}/staging/services.cfg '
            '-c {configs_dir}/staging/placement.cfg '
            '-s 0 -t 2700 -d testTarget -W -u'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

    def test_local_config_rendered_with_same_perms(self):
        phase = DeployerPhase(config='services', local="local.yaml")
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['services'],
            stage='staging')
        utils.touch(os.path.join(workspace.local_dir, "local.yaml"))

        command = self.get_deployer_command_for_run(phase, workspace, 'staging')

        expected_command = (
            'juju-deployer -c {configs_dir}/services '
            '-c {configs_dir}/local.yaml-local.cfg '
            '-s 0 -t 2700 -d testTarget -W -u'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)
        mode = os.stat(os.path.join(workspace.repo_dir, 'local.yaml-local.cfg')).st_mode
        group_other_bits = (stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH |
                            stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP)

        self.assertFalse(mode & group_other_bits,
                         "The rendered local config is not accessible to "
                         "group or others.")

    def test_multiple_local_config_rendered_with_same_perms(self):
        local = "secrets,local.yaml"
        phase = DeployerPhase(config='services', local=local)
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['services'],
            stage='staging')
        utils.touch(os.path.join(workspace.local_dir, "secrets"))
        utils.touch(os.path.join(workspace.local_dir, "local.yaml"))

        command = self.get_deployer_command_for_run(phase, workspace, 'staging')

        expected_command = (
            'juju-deployer -c {configs_dir}/services '
            '-c {configs_dir}/secrets-local.cfg '
            '-c {configs_dir}/local.yaml-local.cfg '
            '-s 0 -t 2700 -d testTarget -W -u'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

        group_other_bits = (stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH |
                            stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP)

        for config in local.split(','):
            path = os.path.join(workspace.repo_dir,
                                '{}-local.cfg'.format(config))
            mode = os.stat(path).st_mode

            self.assertFalse(mode & group_other_bits,
                             "The rendered local config for {} is not "
                             "accessible to group or others.".format(config))

    @mock.patch('mojo.phase.SecretsPhase')
    def test_secrets_automatically_run(self, secrets):
        phase = DeployerPhase(config='services.cfg', target='test')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'services.cfg'],
            stage='staging')

        self.get_deployer_command_for_run(phase, workspace, 'staging')

        secrets.assert_called_with(auto_secrets=True)
        called_methods = [c[0] for c in secrets.mock_calls]
        assert ('().run' in called_methods)

    @mock.patch('mojo.phase.SecretsPhase')
    def test_only_charmstore_charms(self, _secrets):
        phase = DeployerPhase(config='services.cfg', target='test')
        self.assertFalse(phase._only_charmstore_charms(['mojo/tests/testdata/deploy-config-local-charms.yaml']))
        self.assertTrue(phase._only_charmstore_charms(['mojo/tests/testdata/deploy-config-charmstore-charms.yaml']))
        # We should return false (i.e. not *only* charmstore charms) if
        # there's a mix of local and charmstore charms.
        self.assertFalse(phase._only_charmstore_charms(['mojo/tests/testdata/deploy-config-mixed-charms.yaml']))


class BundlePhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    @mock.patch('mojo.phase.bicommand')
    @mock.patch('mojo.phase.BundlePhase._confirm_charm_repo')
    @mock.patch('mojo.phase.BundlePhase._get_target')
    @mock.patch('mojo.juju.Status.check_and_wait')
    @mock.patch('mojo.phase.BundlePhase._get_command')
    def get_bundle_command_for_run(self, phase, workspace, stage, _get_command, _check_wait,
                                   _get_target, _confirm_charm_repo, _bicommand):
        _get_target.return_value = "testTarget"
        _bicommand.return_value = (0, 'fake-output')
        _get_command.return_value = "juju deploy"

        project = Project('unused', 'trusty', '/srv/mojo')
        phase.run(project, workspace, stage)

        return _bicommand.call_args[0][0]

    @mock.patch('mojo.phase.bicommand')
    @mock.patch('mojo.phase.BundlePhase._confirm_charm_repo')
    @mock.patch('mojo.phase.BundlePhase._get_target')
    @mock.patch('mojo.juju.Status.check_and_wait')
    @mock.patch('mojo.phase.BundlePhase._get_command')
    def get_bundle_command_for_diff(self, phase, workspace, stage, _get_command, _check_wait,
                                    _get_target, _confirm_charm_repo, _bicommand):
        _get_target.return_value = "testTarget"
        _bicommand.return_value = (0, 'fake-output')
        _get_command.return_value = "juju"

        project = Project('unused', 'trusty', '/srv/mojo')
        phase.diff(project, workspace, stage)

        return _bicommand.call_args[0][0]

    def test_default_config_rendered_for_bundle(self):
        phase = BundlePhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/bundle', 'bundle'],
            stage='staging')

        command = self.get_bundle_command_for_run(phase, workspace, 'staging')

        # Verify that the configs were both rendered.
        self.assertTrue(os.path.exists(
            '{}/bundle'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/bundle'.format(workspace.repo_dir)))
        # Verify that the configs were included in the juju command.
        expected_command = (
            'juju deploy {configs_dir}/bundle '
            '--overlay {configs_dir}/staging/bundle'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

    def test_multiple_configs_rendered_for_bundle(self):
        phase = BundlePhase(config='bundle.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/bundle.cfg', 'bundle.cfg'],
            stage='staging')

        command = self.get_bundle_command_for_run(phase, workspace, 'staging')

        # Verify the configs were rendered
        self.assertTrue(os.path.exists(
            '{}/bundle.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/bundle.cfg'.format(workspace.repo_dir)))
        # Verify configs are in the juju command
        expected_command = (
            'juju deploy {configs_dir}/bundle.cfg '
            '--overlay {configs_dir}/staging/bundle.cfg'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

    def test_diff_for_bundle(self):
        phase = BundlePhase(config='bundle.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/bundle.cfg', 'bundle.cfg'],
            stage='staging')

        command = self.get_bundle_command_for_diff(phase, workspace, 'staging')

        expected_command = (
            'juju --dry-run {configs_dir}/bundle.cfg '
            '--overlay {configs_dir}/staging/bundle.cfg'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

    def test_comma_separated_configs_rendered_for_bundle(self):
        phase = BundlePhase(config='services.cfg,placement.cfg')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/services.cfg', 'staging/placement.cfg', 'services.cfg'],
            stage='staging')

        command = self.get_bundle_command_for_run(phase, workspace, 'staging')

        # Verify that the configs were rendered
        self.assertTrue(os.path.exists(
            '{}/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/services.cfg'.format(workspace.repo_dir)))
        self.assertTrue(os.path.exists(
            '{}/staging/placement.cfg'.format(workspace.repo_dir)))
        # Verify configs are in the juju command
        expected_command = (
            'juju deploy {configs_dir}/services.cfg '
            '--overlay {configs_dir}/staging/services.cfg '
            '--overlay {configs_dir}/staging/placement.cfg'.format(
                configs_dir=workspace.repo_dir))
        self.assertEqual(expected_command, command)

    @mock.patch('mojo.phase.SecretsPhase')
    def test_secrets_automatically_run(self, secrets):
        phase = BundlePhase(config='bundle.cfg', target='test')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['staging/bundle.cfg', 'bundle.cfg'],
            stage='staging')

        self.get_bundle_command_for_run(phase, workspace, 'staging')

        secrets.assert_called_with(auto_secrets=True)
        called_methods = [c[0] for c in secrets.mock_calls]
        assert ('().run' in called_methods)


class CollectPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def setUp(self):
        self.empty_project = Project('unused', 'trusty', '/tmp')

    @mock.patch('codetree.config.Config')
    def test_codetree_is_called(self, codetree_Config):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')

        called_methods = [c[0] for c in codetree_Config.mock_calls]
        assert ('().build' in called_methods)

    @mock.patch('codetree.config.Config')
    def test_codetree_is_called_with_workers(self, codetree_Config):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        # With no workers specified to start with
        phase.run(self.empty_project, workspace, 'staging')
        called_without_workers = 'call().build()' in \
            [str(c) for c in codetree_Config.mock_calls]
        assert (called_without_workers)

        # And now with workers specified
        phase.options.update({'workers': 2})
        phase.run(self.empty_project, workspace, 'staging')

        called_with_workers = 'call().build(workers=2)' in \
            [str(c) for c in codetree_Config.mock_calls]
        assert (called_with_workers)

    @mock.patch('codetree.config.Config')
    def test_codetree_is_called_with_workers_unsupported(self, codetree_Config):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        phase.options.update({'workers': 2})
        errmsg = "build() got an unexpected keyword argument 'workers'"
        codetree_Config.return_value.build.side_effect = TypeError(errmsg)

        with self.assertRaises(CollectPhaseException):
            phase.run(self.empty_project, workspace, 'staging')

        called_with_workers = 'call().build(workers=2)' in \
            [str(c) for c in codetree_Config.mock_calls]
        assert (called_with_workers)

    @mock.patch('codetree.config.Config')
    def test_codetree_failure_exception(self, _codetree_Config):
        _codetree_Config.return_value.build.return_value = False
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        with self.assertRaises(CollectPhaseException):
            phase.run(self.empty_project, workspace, 'staging')

    @mock.patch('mojo.phase.CharmRepoPhase')
    @mock.patch('codetree.config.Config')
    def test_repo_automatically_runs(self, codetree_Config, charm_repo):
        phase = CollectPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['collect', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')

        called_methods = [c[0] for c in charm_repo.mock_calls]
        assert ('().run' in called_methods)


class NagiosCheckPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    @mock.patch('commands.getstatusoutput')
    def test_run_nagios_checks(self, _get_status_output):
        phase = NagiosCheckPhase()
        # Test a single passing nagios check
        _get_status_output.return_value = 0, (
            "+ /usr/lib/nagios/plugins/check_load -w 0.7,0.7,0.7 -c 1,1,1\n"
            "OK - load average: 0.01, 0.01, 0.00|load1=0.010;0.700;1.000;0; load5=0.010;0.700;1.000;0;\n"
            "load15=0.000;0.700;1.000;0;\n"
            "+ echo MOJO_NAGIOS_OK\n"
            "MOJO_NAGIOS_OK\n"
            )
        expected_output = "", "1 OK on postgresql/0"
        self.assertEqual(phase.run_nagios_checks('postgresql/0'), expected_output)

        # Test one passing, one failing nagios check
        _get_status_output.return_value = 0, (
            "+ /usr/lib/nagios/plugins/check_load -w 0.7,0.7,0.7 -c 1,1,1\n"
            "OK - load average: 0.01, 0.01, 0.00|load1=0.010;0.700;1.000;0; load5=0.010;0.700;1.000;0;\n"
            "load15=0.000;0.700;1.000;0;\n"
            "+ echo MOJO_NAGIOS_OK\n"
            "MOJO_NAGIOS_OK\n"
            "+ /usr/lib/nagios/plugins/check_ksplice.py\n"
            "OK - Effective version 4.4.0-103.126 is older than latest on-disk version 4.4.0-112\n"
            "canonical-livepatch is supported on this host, please use it instead of ksplice!\n"
            "MOJO_NAGIOS_FAIL\n"
            )
        expected_output = (
            "1 FAIL on postgresql/0:\n"
            "+ /usr/lib/nagios/plugins/check_ksplice.py\n"
            "OK - Effective version 4.4.0-103.126 is older than latest on-disk version 4.4.0-112\n"
            "canonical-livepatch is supported on this host, please use it instead of ksplice!"
            ), "1 OK on postgresql/0"
        self.assertEqual(phase.run_nagios_checks('postgresql/0'), expected_output)

        # Test nagios not installed or nagios checks not found (this can
        # have a zero or non-zero return code, so we'll test both).
        _get_status_output.return_value = 0, 'grep: /etc/nagios/nrpe.d/*: No such file or directory'
        expected_output = "Nagios not installed, or no nagios checks found on postgresql/0", ""
        self.assertEqual(phase.run_nagios_checks('postgresql/0'), expected_output)

        # Test nagios user not found (with non-zero return code).
        _get_status_output.return_value = 1, 'sudo: unknown user: nagios'
        expected_output = "Nagios not installed, or no nagios checks found on postgresql/0", ""
        self.assertEqual(phase.run_nagios_checks('postgresql/0'), expected_output)

        # Test unknown error
        _get_status_output.return_value = 1, "We don't recognise this error"
        with self.assertRaises(NagiosCheckPhaseException):
            phase.run_nagios_checks('postgresql/0')

    @mock.patch('mojo.phase.NagiosCheckPhase.run_nagios_checks')
    @mock.patch('mojo.juju.Status')
    def test_all_checks_dispatched(self,  _juju_status, _run_nagios_checks):
        """Test that all expected checks are dispatched."""
        # Side effect that returns specific lists of machines for each service
        def smn_side_effect(service):
            machine_numbers = {'foo': [1, 2],
                               'bar': [4],
                               'baz': [5]}
            return machine_numbers[service]

        # Side effect that returns specific units for each service
        def su_side_effect(service):
            service_units = {'foo': ['foo/0', 'foo/1'],
                             'bar': ['bar/2'],
                             'baz': ['baz/1']}
            return service_units[service]

        # Set up all the mockery
        _juju_status().services_list.return_value = ['foo', 'bar', 'baz']
        _juju_status().service_machine_numbers.side_effect = smn_side_effect
        _juju_status().service_units.side_effect = su_side_effect
        _run_nagios_checks.return_value = "", "Everything OK"
        # Define a minimal, empty project and phase with 2 services to check
        self.empty_project = Project('unused', 'trusty', '/tmp')
        phase = NagiosCheckPhase()
        phase.options.update({'services': 'foo,bar'})
        workspace = self.create_workspace()
        self.create_spec(
                path=workspace.spec_dir,
                configs=['collect', 'services.cfg'], stage='staging')
        # Run the phase
        phase.run(self.empty_project, workspace, 'staging')
        # Check expected tests were all dispatched, as seen by calls to
        # run_nagios_checks
        all_skip = ['check_etc_bzr', 'check_log_archive_status']
        _run_nagios_checks.assert_has_calls([
            mock.call('foo/0', all_skip),
            mock.call('foo/1', all_skip),
            mock.call('bar/2', all_skip),
            ])

    @mock.patch('mojo.phase.ThreadPoolExecutor')
    @mock.patch('mojo.juju.Status')
    def test_max_worker_calculation(self,  _juju_status, _executor):
        """Test that max_workers is properly calculated."""
        # Side effect that returns specific lists of machines for each service
        def smn_side_effect(service):
            machine_numbers = {'foo': [1, 2],
                               'bar': [4],
                               'baz': [5],
                               'nada': None}
            return machine_numbers[service]

        # Set up all the mockery
        _juju_status().services_list.return_value = ['foo', 'bar', 'baz', 'nada']
        _juju_status().service_machine_numbers.side_effect = smn_side_effect
        # Define a minimal, empty project and phase with 2 services to check
        self.empty_project = Project('unused', 'trusty', '/tmp')
        phase = NagiosCheckPhase()
        phase.options.update({'services': 'foo,bar'})
        workspace = self.create_workspace()
        self.create_spec(
                path=workspace.spec_dir,
                configs=['collect', 'services.cfg'], stage='staging')
        # Run the phase
        phase.run(self.empty_project, workspace, 'staging')
        # Check call to ThreadPoolExecutor for max_workers value
        # the value is "3", for the sum of machines in the "foo" and "bar"
        # services.
        _executor.assert_called_once_with(max_workers=3)


class CharmAuditPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def test_parse_charm_audit_output(self):
        # In this example, landscape-client and ntp don't have a codetree-collect-info.yaml
        # file, but telegraf does.
        audit_output = dedent("""\
            landscape-client-1:
              application: landscape-client
            ntp-0:
              application: ntp
            telegraf-12:
              application: telegraf
              collect_date: '2017-11-01 11:34:44.479841'
              collect_url: git://git.launchpad.net/telegraf-charm;revno=ad93f7333da35a
              hashes:
                COPYING.txt: 8ceb4b9ee5adedde47b31e975c1d90c73ad27b6b165a1dcd80c7c545eb65b903
            """)
        expected_output = {
            "landscape-client": {"Unknown": ["landscape-client-1"]},
            "ntp": {"Unknown": ["ntp-0"]},
            "telegraf": {"git://git.launchpad.net/telegraf-charm;revno=ad93f7333da35a": ["telegraf-12"]},
            }
        phase = CharmAuditPhase()
        self.assertEqual(phase.parse_charm_audit_output(audit_output), expected_output)

    def test_charm_audit_summary(self):
        # In this example, landscape-client and ntp don't have a codetree-collect-info.yaml
        # file, but telegraf does.
        parsed_audit_output = {
            "landscape-client": {"Unknown": ["landscape-client-1"]},
            "ntp": {"Unknown": ["ntp-0"]},
            "telegraf": {"git://git.launchpad.net/telegraf-charm;revno=ad93f7333da35a": ["telegraf-12"]},
            }
        expected_output = dedent("""\
            Application: landscape-client
              Unable to determine charm version for landscape-client-1
            Application: ntp
              Unable to determine charm version for ntp-0
            Application: telegraf
              The charm git://git.launchpad.net/telegraf-charm;revno=ad93f7333da35a is deployed on telegraf-12
            """).strip()
        phase = CharmAuditPhase()
        self.assertEqual("\n".join(phase.charm_audit_summary(parsed_audit_output)), expected_output)


class CharmRepoPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def setUp(self):
        self.empty_project = Project('unused', 'trusty', '/tmp')

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_repo_default(self, build_all, build_from_config):
        phase = CharmRepoPhase()
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')

        build_from_config.assert_called_with(
            os.path.join(workspace.spec_dir, 'repo'),
            os.path.join(
                os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_repo_with_config(self, build_all, build_from_config):
        phase = CharmRepoPhase(config='repo-update')
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo-update', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')

        build_from_config.assert_called_with(
            os.path.join(workspace.spec_dir, 'repo-update'),
            os.path.join(
                os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')

    @mock.patch('mojo.charm_repo.CharmRepo.build_from_config')
    @mock.patch('mojo.charm_repo.CharmRepo.build_all')
    def test_auto_repo(self, build_all, build_from_config):
        phase = CharmRepoPhase(auto_repo=True)
        workspace = self.create_workspace()
        self.create_spec(
            path=workspace.spec_dir,
            configs=['repo', 'services.cfg'], stage='staging')

        phase.run(self.empty_project, workspace, 'staging')
        build_all.assert_called_with(
            os.path.join(os.path.dirname(workspace.spec_dir), 'build'),
            'trusty')


class SecretsPhaseTestCase(TestCase, utils.MojoTestCaseMixin):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp(prefix="mojo-test-secrets")
        self.empty_project = Project('unused', 'xenial', self.tempdir)

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @mock.patch('logging.warn')
    def test_no_secrets_dir(self, _warn):
        # If we're automatically running the phase as part of collect, for
        # instance, we only expect a warning to be logged if the secrets
        # directory doesn't exist.
        phase = SecretsPhase(auto_secrets=True)
        workspace = self.create_workspace()
        phase.run(self.empty_project, workspace, 'staging')
        _warn.assert_called_with("Automatic secrets phase ran but secrets directory {}/LOCAL/unused/staging does "
                                 "not exist!".format(self.tempdir))

        # If we're calling the phase without auto_secrets, we expect an
        # exception if the secrets directory doesn't exist.
        phase = SecretsPhase()
        with self.assertRaises(FileDoesNotExistException):
            phase.run(self.empty_project, workspace, 'staging')

    def test_secrets_updates(self):
        # First we want to create our secrets directory and populate it with a
        # file.
        stage = 'staging'
        secrets_dir = os.path.join(self.empty_project.mojo_root, 'LOCAL', self.empty_project.name, stage)
        os.makedirs(secrets_dir)
        with open(os.path.join(secrets_dir, 'test-file'), 'w') as f:
            f.write("test content")

        # Now run our secrets phase.
        phase = SecretsPhase()
        workspace = self.create_workspace()

        # Check the workspace local directory is empty before we run the
        # secrets phase.
        self.assertEqual(os.listdir(workspace.local_dir), [])
        phase.run(self.empty_project, workspace, stage)

        # Now check the workspace local directory has the file we expect.
        self.assertEqual(os.listdir(workspace.local_dir), ['test-file'])
        with open(os.path.join(workspace.local_dir, 'test-file'), 'r') as f:
            self.assertEqual(f.read(), "test content")

        # Now change the file in the workspace local directory to read-only
        # and confirm we can still overwrite it from our secrets directory.
        os.chmod(os.path.join(workspace.local_dir, 'test-file'), 0o400)
        # Change the contents of the file in the secrets directory so we can
        # look for this changed content after running the secrets phase.
        with open(os.path.join(secrets_dir, 'test-file'), 'w') as f:
            f.write("read-only file has been overwritten")

        # Run the secrets phase again and confirm our read-only file has
        # been updated.
        phase.run(self.empty_project, workspace, stage)
        self.assertEqual(os.listdir(workspace.local_dir), ['test-file'])
        with open(os.path.join(workspace.local_dir, 'test-file'), 'r') as f:
            self.assertEqual(f.read(), "read-only file has been overwritten")
