# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
import errno
import grp
import logging
import os
import pickle
import pwd
import select
import shutil
import subprocess
import sys
import time

from hashlib import sha1
from .exceptions import Unprivileged


class CacheDisk():
    hits = 0
    misses = 0
    stores = 0
    deletes = 0

    @classmethod
    def get_stats(cls):
        try:
            cls.hit_rate = float(cls.hits) / (cls.hits + cls.misses) * 100
        except ZeroDivisionError:
            cls.hit_rate = None
        return ('hit_rate={0.hit_rate:.4} hits={0.hits} misses={0.misses}'
                'stores={0.stores} deltes={0.deletes}'.format(cls))

    @classmethod
    def ttl(cls, new_value=None):
        '''Return current CACHE_DISK_TTL (saved to environ), set if if value
           passed
        '''
        if new_value:
            os.environ["CACHE_DISK_TTL"] = str(new_value)
        # Disabled by default
        return int(os.environ.get("CACHE_DISK_TTL", 0))

    @classmethod
    def dir(cls, new_value=None):
        '''Return current CACHE_DISK_DIR (saved to environ), set if if value
           passed
        '''
        if new_value:
            os.environ["CACHE_DISK_DIR"] = str(new_value)
        return os.environ.get("CACHE_DISK_DIR", "~/.cache/mojo")

    @classmethod
    def delete_cache_dir(cls, obj, unique_attrs):
        cache_dir = os.path.expanduser(CacheDisk.dir())
        attrs_dict = [{k: obj.__dict__[k]} for k in unique_attrs]
        cache_key_dir = os.path.join(cache_dir,
                                     sha1(str(attrs_dict)).hexdigest())
        debug_msg = "attrs_dict={} cache_key_dir={}".format(
            attrs_dict, cache_key_dir)
        logging.debug("{}:   DEL: {}".format(cls.__name__, debug_msg))
        cls.deletes += 1
        shutil.rmtree(cache_key_dir, ignore_errors=True)

    @classmethod
    def delete(cls, unique_attrs=()):
        def do_delete(func):
            def wrap(*args, **kwargs):
                # Only delete if ttl!=0 (ie use caching)
                if CacheDisk.ttl():
                    cls.delete_cache_dir(args[0], unique_attrs)
                return func(*args, **kwargs)
            return wrap
        return do_delete

    @classmethod
    def cache(cls, unique_attrs=()):
        '''Simple file based cache decorator for class methods
           unique_attrs: object attributes to use hen building the unique key,
                         together with function *args and **kwargs.

        Sample usage:
            @CacheDisk.deco(unique_attrs=('attr1', 'attr2',))
            def a_function(foo, bar, baz=None):
                    ...
        '''
        def do_cache(func):
            def wrap(*args, **kwargs):
                cache_dir = os.path.expanduser(CacheDisk.dir())
                # part of the key that identifies this function
                key_func = "{}.{}".format(func.__module__, func.__name__)
                # part of the key that identifies this object
                attrs_dict = [{k: args[0].__dict__[k]} for k in unique_attrs]
                # cache_key_dir holds all cached values for this object
                # so that it's easy to delete an object's cache by
                # rmtree(cache_key_dir)
                cache_key_dir = sha1(str(attrs_dict)).hexdigest()
                # part of the key that identifies calling args
                call_args = {
                    "args": args[1:],  # skip args[0] (the object id / self)
                    "kwargs": kwargs,
                    }
                cache_key_file = "{}.{}".format(
                    key_func, sha1(str(call_args)).hexdigest())
                cache_key_dir = os.path.join(cache_dir, cache_key_dir)
                filepath = os.path.join(cache_key_dir, cache_key_file)
                debug_msg = "attrs_dict={} func={} call_args={}".format(
                    attrs_dict, key_func, call_args)

                # verify cached object exists and didn't expire if ttl!=0
                ttl = CacheDisk.ttl()
                if ttl:
                    if os.path.exists(filepath):
                        mtime = os.path.getmtime(filepath)
                        age = time.time() - mtime
                        if age < ttl:
                            value = pickle.load(open(filepath, "rb"))['value']
                            logging.debug("{}:  HIT: {} value={} from {}"
                                          "".format(cls.__name__,
                                                    debug_msg, value,
                                                    filepath))
                            cls.hits += 1
                            return value
                    logging.debug("{}:  MISS: {}"
                                  "".format(cls.__name__, debug_msg))
                    cls.misses += 1

                # cache miss, call the original function
                value = func(*args, **kwargs)

                # save result for next time only if ttl!=0 (avoid creating
                # cache if not asked)
                if ttl:
                    if not os.path.exists(cache_key_dir):
                        os.makedirs(cache_key_dir)
                        open(os.path.join(cache_key_dir, "key.txt"),
                             "w").write("{}\n".format(attrs_dict))
                    # save call_args alongside value, for debugging
                    # or textual-grepping purposes
                    to_cache = {'value': value, 'debug_msg': debug_msg, }
                    logging.debug("{}: STORE: {} value={} to {}"
                                  "".format(cls.__name__, debug_msg, value,
                                            filepath))
                    cls.stores += 1
                    pickle.dump(to_cache, open(filepath, "wb"))
                return value
            return wrap
        return do_cache


def setup_logging(log_level):
    logFormatter = logging.Formatter(
        fmt="%(asctime)s [%(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S")
    rootLogger = logging.getLogger()
    rootLogger.setLevel(log_level)

    consoleHandler = logging.StreamHandler()
    consoleHandler.setFormatter(logFormatter)
    rootLogger.addHandler(consoleHandler)


def mkdirs(path, user=os.environ.get('USER'), group=None, mode=0o755):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        elif e.errno == errno.EACCES:
            raise OSError("Permission denied attempting to create directory "
                          "{}".format(path))
        else:
            raise OSError("Creation of directory failed: {}".format(path))
    if user:
        uid = pwd.getpwnam(user).pw_uid
    else:
        uid = os.getuid()
    if group:
        gid = grp.getgrnam(group).gr_gid
    else:
        gid = pwd.getpwuid(uid).pw_gid
    try:
        os.chown(path, uid, gid)
    except OSError as e:
        raise OSError("Unable to chown {} as uid: {}, gid: {}".format(path,
                                                                      uid,
                                                                      gid))


def mksymlink(source, link_name):
    if os.path.exists(link_name):
        if os.path.islink(link_name):
            if os.path.realpath(source) == os.path.realpath(link_name):
                return True
            else:
                raise OSError("{} link exists but its target {} does not "
                              "match {}".format(link_name,
                                                os.path.realpath(link_name),
                                                os.path.realpath(source)))
        else:
            raise OSError("{} exists but is not a symlink".format(link_name))
    else:
        os.symlink(source, link_name)
        return True


def is_config_line(line):
    """Checks if a line is has content and isn't a comment"""
    if line.strip() and not line.strip().startswith("#"):
        return True
    return False


def get_dep_dir(depname):
    deps_dir = os.environ.get("MOJO_DEPS_DIR",
                              os.path.join(os.path.dirname(__file__),
                                           '..', 'deps'))
    dep_dir = os.path.join(os.path.abspath(deps_dir), depname)
    if os.path.isdir(dep_dir):
        return dep_dir
    else:
        return None


def get_dep_prog(progname):
    locations = os.environ.get("PATH").split(os.pathsep)
    for location in locations:
        candidate = os.path.join(location, progname)
        if os.path.isfile(candidate):
            return candidate
    deps_dir = os.environ.get("MOJO_DEPS_DIR",
                              os.path.join(os.path.dirname(__file__),
                                           '..', 'deps'))
    prog = os.path.join(os.path.abspath(deps_dir), progname, progname)
    if os.path.isfile(prog):
        return prog
    else:
        return None


# This meant as a replacement for commands.getstatusoutput.  By default, it
# works more or less identically.  However, if you pass showoutput=True as
# its second parameter, it will show the process of your command as it goes.
# This is useful for long-running commands whose output is interesting to
# watch as well as to programatically analyse once it's finished.
# Stolen mercilessly from https://gist.github.com/maw/1711614 (with thanks)
def bicommand(command, showoutput=False, read_bytes=1, env=None):

    if showoutput not in [True, False]:
        raise "showoutput takes a boolean argument only"

    pipe = subprocess.Popen(command, shell=True,
                            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, env=env)

    output = ""
    while True:
        to_read, _, _ = select.select([pipe.stdout, pipe.stderr], [], [], 0.1)
        if len(to_read) == 0:
            continue

        eof = True
        for stream in to_read:
            line = stream.read(read_bytes)
            if line != "":
                eof = False
            output += line
            if showoutput:
                sys.stdout.write(line)
                pass

        if eof:
            break

        pass

    if output[-1:] == '\n':
        output = output[:-1]
        pass

    status = pipe.wait()

    return status, output


# Decorator for methods that must be run as root
def needs_root(method):
    def wrapped(*args, **kwargs):
        if os.geteuid() != 0:
            raise Unprivileged("This function must be run as a privileged "
                               "user")
        return method(*args, **kwargs)
    return wrapped


# Decorator for methods that should warn if run as root
def warn_if_root(method):
    def wrapped(*args, **kwargs):
        if os.geteuid() == 0 and not args[0].break_everything:
            print("This should not be run as root, exiting.\n"
                  "Possibly use --break-everything option "
                  "if you are really sure!")
            exit
        else:
            return method(*args, **kwargs)
    return wrapped


class chdir(object):
    """Change cwd temporarily."""
    def __init__(self, path):
        self.old_dir = os.getcwd()
        self.new_dir = path

    def __enter__(self):
        os.chdir(self.new_dir)

    def __exit__(self, *args):
        os.chdir(self.old_dir)


def secs2human(seconds):
    ''' Convert seconds to human readable (e.g. 1h 32m 10s) or return 'None'
    on invalid input and let the calling function deal with it.
    '''
    if seconds is True or seconds is False:
        return None
    try:
        secs = int(round(float(seconds)))
    except (TypeError, ValueError):
        return None
    if secs == 0:
        return '0s'
    hours = secs // 3600
    minutes = secs // 60 % 60
    seconds = secs % 60
    output = ''
    if hours > 0:
        output += '{}h'.format(hours)
    if minutes > 0:
        if len(output) > 0:
            output += ' '
        output += '{}m'.format(minutes)
    if seconds > 0:
        if len(output) > 0:
            output += ' '
        output += '{}s'.format(seconds)
    return output
