# Copyright 2014 Canonical Ltd.  This software is licensed under the
# GNU General Public License version 3 (see the file LICENSE).
"""A helper for attaching volumes to Juju instances"""
from __future__ import print_function
from subprocess import check_call
import os
import sys

import cinderclient.client
import cinderclient.exceptions

from mojo.cli import MojoArgumentParser
from mojo.juju import Status
from mojo.juju.utils import get_juju_command
from mojo.utils import is_config_line


class VolumeAttachError(Exception):
    pass


class VolumeConfigError(Exception):
    pass


class Volume(object):
    def __init__(self, nova_volume):
        super(Volume, self).__init__()
        self._volume = nova_volume

    def attached_unit(self, service):
        "Unit ID this volume is attached to"
        server = self.attached_machine()
        if not server:
            return None
        status = Status()
        units = status.yaml_status()['services'].get(service, {}).get('units', {})
        if not units:
            return None
        for (mid, mattrs) in status.yaml_status()['machines'].items():
            if mattrs['instance-id'] == server:
                for uid, uattrs in units.items():
                    if uattrs['machine'] == mid:
                        return uid
        return None

    def attached_machine(self):
        "Machine ID this volume is attached to"
        if self._volume.attachments:
            return self._volume.attachments[0]['server_id']
        else:
            return None

    @property
    def attached(self):
        "Is this volume attached?"
        if self.attached_machine():
            return True
        return False

    def attach(self, machine_id):
        # This works for the current (buggy) openstack implementation, as it's
        # completely ignored. In the future, we may need to calculate a device
        # name
        device = '/dev/vdz'
        cmd = ['nova', 'volume-attach', machine_id, self.volid, device]
        with open(os.devnull, 'w') as devnull:
            check_call(cmd, stdout=devnull, stderr=devnull)

    @property
    def volid(self):
        return self._volume.id


class VolumeMap(object):
    def __init__(self):
        super(VolumeMap, self).__init__()
        if os.environ['OS_REGION_NAME']:
            region_name = os.environ['OS_REGION_NAME']
        else:
            region_name = None
        self.cinder = cinderclient.client.Client(1,
                                                 os.environ['OS_USERNAME'],
                                                 os.environ['OS_PASSWORD'],
                                                 os.environ['OS_TENANT_NAME'],
                                                 os.environ['OS_AUTH_URL'],
                                                 region_name=region_name)

    def volume_by_name(self, volume_name):
        for volume in self.cinder.volumes.list():
            if volume.display_name == volume_name:
                return Volume(volume)
        return None

    def volume_by_id(self, volume_id):
        try:
            volume = self.cinder.volumes.get(volume_id)
        except cinderclient.exceptions.NotFound:
            raise VolumeConfigError("Volume id {} not found".format(volume_id))
        return Volume(volume)

    def volumes(self):
        return [Volume(v) for v in self.cinder.volumes.list()]

    def mapped_units(self, service, volume_ids):
        "Return a map of units to the given volumes, plus a list of volumes "
        "not attached"
        volumes = [self.volume_by_id(v) for v in volume_ids]
        unit_map = {}
        for volume in volumes:
            if not volume:
                raise KeyError("No such volume: {}".format(volume.volid))
            # is the volume already attached?
            if volume.attached:
                # is it attached to a unit of the named service?
                unit = volume.attached_unit(service)
                if unit:
                    unit_map[volume.volid] = unit
                else:
                    raise VolumeAttachError("Volume {} is attached, but not "
                                            "to service {}"
                                            "".format(volume.volid, service))
            else:
                unit_map[volume.volid] = None

        return unit_map

    def attach_volumes(self, service, volume_ids):
        "Attach any unattached volumes to units of the given service"
        unit_map = self.mapped_units(service, volume_ids)
        status = Status()
        for volume_id in [v for (v, u) in unit_map.items() if u is None]:
            if status.service_units(service):
                units = [u for u in status.service_units(service)
                         if u not in unit_map.values()]
            else:
                raise VolumeConfigError("No service {} found".format(service))
            if units:
                the_unit = units[0]
                machine_num = status.yaml_status()['services'][service]['units'][the_unit]['machine']
                machine_id = status.machine_instance_id(machine_num)
                self.volume_by_id(volume_id).attach(machine_id)
                unit_map[volume_id] = the_unit
            else:
                raise VolumeAttachError("Not enough units to attach all "
                                        "volumes.")

        return unit_map


def process_config(filename):
    config = []
    with open(filename) as conf:
        lineno = 0
        for line in conf:
            if is_config_line(line):
                lineno += 1
                opts = line.strip().split()
                if not len(opts) >= 2:
                    raise VolumeConfigError("Invalid config "
                                            "{} at line {}".format(filename,
                                                                   lineno))
                config.append(opts)
    return config


def main():
    ap = MojoArgumentParser(description="Attach volumes to service units",
                            epilog="Specify either a config or other options, "
                                   "not both")
    ap.add_argument("--config", nargs=1, help="configuration file")
    ap.add_argument("--service", nargs=1, help="service name")
    ap.add_argument("--volumes", nargs="*", help="list of volumes")

    args = ap.parse_args()
    if args.config:
        if args.service or args.volumes:
            print("Specify either a config or other options, not both",
                  file=sys.stderr)
            sys.exit(1)
        config = process_config(args.config[0])
        for opts in config:
            service, volumes = opts[0], opts[1:]
            vmap = VolumeMap()
            unit_map = vmap.attach_volumes(service, volumes)
            cmd = ['juju', get_juju_command('set'), service, 'volume-ephemeral-storage=False']
            volume_map = {}
            for (vol_id, serv_id) in unit_map.items():
                volume_map[serv_id] = vol_id.encode('ascii', 'ignore')
            cmd.append('volume-map={}'.format(volume_map))
            check_call(cmd)

    else:
        if not args.service and args.volumes:
            ap.print_help(file=sys.stderr)
            sys.exit(1)
        vmap = VolumeMap()
        print(vmap.attach_volumes(args.service[0], args.volumes))
