import logging
import os
import shutil
from mojo.exceptions import ArgumentError
from mojo.spec import Spec


class NoSuchWorkspaceError(Exception):
    pass


class Workspace(object):
    "A mojo workspace"
    def __init__(self, name, path, stage=None, spec_url=None, manifest_file=None, logfile=None):
        if name is None:
            raise ArgumentError("Workspace setting missing. Use --workspace or set MOJO_WORKSPACE")
        self.name = name
        self.manifest_file = manifest_file
        if not path:
            self.path = os.path.abspath(os.curdir)
            logging.info("No workspace path specified, using {}"
                         "".format(self.path))
        else:
            self.path = os.path.abspath(path)
        if not os.path.exists(self.path):
            raise NoSuchWorkspaceError(path)
        self.stage = stage
        self.spec_url = spec_url
        self.logfile = logfile
        self._spec = None

    @property
    def spec(self):
        if self._spec is None:
            self._spec = Spec(self.spec_dir, self.spec_url, stage=self.stage, manifest_file=self.manifest_file)
        return self._spec

    @property
    def all_dirs(self):
        return [self.path,
                self.build_dir,
                self.local_dir,
                self.log_dir,
                self.spec_dir,
                ]

    def prepare(self):
        "Set up the workspace. Returns True on success, False on failure."
        for path in self.all_dirs:
            if path == self.spec_dir:  # spec dir is initialized next
                continue
            if not os.path.isdir(path):
                try:
                    os.makedirs(path)
                except OSError as e:
                    logging.error("{} ({})".format(e.strerror, e.filename))
                    return False
        self.spec  # Initializes spec
        self.create_file_logger()
        return True

    def clean(self):
        "Reset the workspace"
        map(shutil.rmtree, (d for d in self.all_dirs if os.path.exists(d)))
        return self.prepare()

    def delete(self):
        "Destroy the workspace"
        realpath = os.path.realpath(self.path)
        shutil.rmtree(realpath)

    def create_file_logger(self):
        "If we haven't already, attach a file logger"
        if logging.getLogger('file').handlers:
            return

        if not self.logfile:
            self.logfile = os.path.join(self.log_dir, "mojo.log")

        try:
            logFormatter = logging.Formatter(
                fmt="%(asctime)s [%(levelname)s] %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S")

            rootLogger = logging.getLogger()
            fileLogger = logging.getLogger('file')
            # If we send output to the file logger specifically, don't propagate it
            # to the root logger as well to avoid duplicate output. So if we want
            # to only send logging output to the file, you would do this:
            #  logging.getLogger('file').info("message for logfile only")
            # rather than this:
            #  logging.info("message for console and logfile")
            fileLogger.propagate = False

            fileHandler = logging.FileHandler(self.logfile)
            fileHandler.setFormatter(logFormatter)
            rootLogger.addHandler(fileHandler)
            fileLogger.addHandler(fileHandler)
        except IOError:
            logging.error("Unable to write to logfile: {}".format(self.logfile))

    @property
    def build_dir(self):
        return os.path.join(self.path, "build")

    @property
    def repo_dir(self):
        return os.path.join(self.path, "charms")

    @property
    def local_dir(self):
        return os.path.join(self.path, "local")

    @property
    def log_dir(self):
        if self.logfile:
            return os.path.dirname(os.path.abspath(self.logfile))
        return os.path.join(self.path, "log")

    @property
    def spec_dir(self):
        return os.path.join(self.path, "spec")
